#!/usr/bin/env python
"""Permutation"""
# standard library
import json
import logging
import os
import random
import re
import sys
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

try:
    # standard library
    import sqlite3
except ImportError:  # pragma: no cover
    pass  # sqlite3 is only required for local development

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.app_config.layout_json import LayoutJson
from tcex.app_config.models.install_json_model import ParamsModel
from tcex.app_config.models.layout_json_model import OutputsModel, ParametersModel
from tcex.backports import cached_property
from tcex.pleb.none_model import NoneModel

# get tcex logger
tcex_logger = logging.getLogger('tcex')


class InputModel(ParamsModel):
    """Input Model"""

    value: Any

    def __hash__(self):
        """Make model hashable."""
        return hash(self.name)


class Permutation:
    """Permutations Module"""

    def __init__(self, logger: Optional[logging.Logger] = None):
        """Initialize Class properties"""
        self.log = logger or tcex_logger

        # properties
        self._input_names = None
        self._input_table = 'inputs'
        self._input_permutations = None
        self._output_permutations = None
        self.fqfn = Path(os.getcwd(), 'permutations.json')
        self.ij = InstallJson(logger=self.log)
        self.lj = LayoutJson(logger=self.log)

    @staticmethod
    def _create_input_model(ij_param: ParamsModel, value: any) -> InputModel:
        """Create an input model from the install.json param model."""
        _input_model = InputModel(**ij_param.dict())
        # manually adding List values due to bug where data is not getting loaded into model in init
        _input_model.intel_type = ij_param.intel_type
        _input_model.playbook_data_type = ij_param.playbook_data_type
        _input_model.valid_values = ij_param.valid_values
        _input_model.value = value
        return _input_model

    @cached_property
    def _display_keywords(self) -> List[str]:
        """Return the display keywords."""
        _keywords = set()
        for param in self.lj.model.params.values():
            for keyword in (param.display or '').split(' '):
                keyword = re.sub(r'[^a-zA-Z0-9_]', '', keyword)
                if keyword in self.ij.model.param_names:
                    _keywords.add(keyword)
        self.log.debug(f'keywords={_keywords}')
        return _keywords

    def _gen_permutations(self, index: Optional[int] = 0, params: Optional[list] = None):
        """Iterate recursively over layout.json parameter names to build permutations.

        .. NOTE:: Permutations are for layout.json based Apps.

        Args:
            index: The current index position in the layout names list.
            params: The current list of args.
        """
        params = params or []
        try:
            # grab the name using the index value of all params in the layout.json file.
            # after the last name is hit the IndexError will trigger collecting outputs.
            name = list(self.lj.model.param_names)[index]

            # get layout.json param name and data
            lj_param: Union[NoneModel, ParametersModel] = self.lj.model.get_param(name)

            # get install.json param to match layout.json param
            ij_param: Union[NoneModel, ParamsModel] = self.ij.model.get_param(name)
            if not isinstance(ij_param, ParamsModel):  # pragma: no cover
                self.handle_error(f'No param found in install.json for "{name}".')

            if self.validate_layout_display(self._input_table, lj_param.display) or ij_param.hidden:
                # only process params that match display query or are hidden
                if name != 'tc_action' and name not in self._display_keywords:
                    params.append(self._create_input_model(ij_param, None))

                    # recursively call method to get all permutations
                    self._gen_permutations(index + 1, list(params))
                elif ij_param.type.lower() == 'boolean':
                    for val in [True, False]:
                        params.append(self._create_input_model(ij_param, val))

                        # update the data in the sqlite db so for next iteration
                        self.db_update_record(self._input_table, name, val)

                        # recursively call method to get all permutations
                        self._gen_permutations(index + 1, list(params))

                        # remove the previous arg before next iteration
                        params.pop()
                elif ij_param.type.lower() in ['choice', 'editchoice']:
                    for val in self.ij.expand_valid_values(ij_param.valid_values):
                        params.append(self._create_input_model(ij_param, val))

                        # update the data in the sqlite db so for next iteration
                        self.db_update_record(self._input_table, name, val)

                        # recursively call method to get all permutations
                        self._gen_permutations(index + 1, list(params))

                        # remove the previous arg before next iteration
                        params.pop()
                else:
                    params.append(self._create_input_model(ij_param, None))

                    # recursively call method to get all permutations
                    self._gen_permutations(index + 1, list(params))
            else:
                # do not add param since it's not required for this permutation
                self._gen_permutations(index + 1, list(params))
        except IndexError:
            # when IndexError is reached all params has been processed
            self._input_permutations.append(params)
            outputs = []

            # iterate of InstallJsonModel -> PlaybookModel -> OutputVariablesModel
            for o in self.ij.model.playbook.output_variables:

                # get layout.json param to match install.json output variable
                lj_output: Union[NoneModel, OutputsModel] = self.lj.model.get_output(o.name)
                if isinstance(lj_output, OutputsModel):
                    valid = self.validate_layout_display(self._input_table, lj_output.display)
                    if lj_output.display is None or not valid:
                        continue
                # output meet permutation check
                outputs.append(o)
            self._output_permutations.append(outputs)

    @property
    def _params_data(self) -> Tuple[str, 'ParamsModel']:
        """Return all defined params from layout.json/install.json, including hidden params."""
        # using inputs from layout.json since they are required to be in order
        # (display field can only use inputs previously defined)
        for input_name in self.lj.model.params:
            # get data from install.json based on name
            ij_data = self.ij.model.get_param(input_name)
            yield ij_data

        # hidden fields will not be in layout.json so they need to be include manually
        for input_name, ij_data in self.ij.model.filter_params(hidden=True).items():
            yield ij_data

    @cached_property
    def action_configurations(self) -> dict:
        """Return action configuration."""
        self.init_permutations()

        _action_configurations = {}
        for index, inputs in enumerate(self._input_permutations):
            for input_ in inputs:
                if input_.name == 'tc_action':
                    action = input_.value
                    _action_configurations.setdefault(action, {'inputs': [], 'outputs': []})
                    _action_configurations[action]['inputs'].extend(inputs)
                    _action_configurations[action]['outputs'].extend(
                        self._output_permutations[index]
                    )
                    break

        for action, data in _action_configurations.items():
            _action_configurations[action] = {
                'inputs': sorted(list(set(data['inputs'])), key=lambda x: x.name),
                'outputs': sorted(list(set(data['outputs'])), key=lambda x: x.name),
            }

        return _action_configurations

    @cached_property
    def db_conn(self) -> 'sqlite3.Connection':
        """Create a temporary in memory DB and return the connection."""
        try:
            return sqlite3.connect(':memory:')
        except sqlite3.Error as e:  # pragma: no cover
            self.handle_error(e)

        return None

    def db_create_table(self, table_name: str, columns: List[str]):
        """Create a temporary DB table.

        Args:
            table_name: The DB table name.
            columns: The DB column names.
        """
        columns = ', '.join([f'''"{c.strip('"').strip("'")}" text''' for c in set(columns)])
        self.log.debug(f'action=db-create-table, table-name={table_name}, columns={columns}')
        sql = f'CREATE TABLE IF NOT EXISTS {table_name} ({columns});'
        try:
            cr = self.db_conn.cursor()
            cr.execute(sql)
        except sqlite3.Error as e:  # pragma: no cover
            self.handle_error(f'SQL create db failed - SQL: "{sql}", Error: "{e}"')

    def db_drop_table(self, table_name: str):
        """Drop a DB table.

        Args:
            table_name: The DB table name.
        """
        sql = f'DROP TABLE IF EXISTS {table_name};'
        try:
            cr = self.db_conn.cursor()
            cr.execute(sql)
        except sqlite3.Error as e:  # pragma: no cover
            self.handle_error(f'SQL drop db failed - SQL: "{sql}", Error: "{e}"')

    def db_insert_record(self, table_name: str, columns: List[str]):
        """Insert records into DB.

        A single row will all values as None so that values can be updated one at a
        time during parsing. The row and values will be used to determine permutations.

        Args:
            table_name: The DB table name.
            columns: The DB column names.
        """
        # sort and unique columns
        columns = sorted(list(set(columns)))
        bindings = ', '.join(['?'] * len(columns))
        columns_string = ', '.join([f'''"{c.strip('"').strip("'")}"''' for c in columns])
        values = [None] * len(columns)
        try:
            sql = f'''INSERT INTO {table_name} ({columns_string}) VALUES ({bindings})'''
            cur = self.db_conn.cursor()
            cur.execute(sql, values)
        except sqlite3.OperationalError as ex:  # pragma: no cover
            self.handle_error(f'SQL insert failed - SQL: "{sql}", Error: "{ex}"')

    def db_update_record(self, table_name: str, column: str, value: str):
        """Update a single column in the row-column create in db_insert_record.

        Args:
            table_name: The DB table name.
            column: The DB column names.
            value: The DB values to store (row data).
        """
        # escape any single quotes in value
        if isinstance(value, str):
            value = value.replace('\'', '\\')
        elif isinstance(value, bool):
            # core expects true/false so we convert bool value to string and lower
            value = str(value).lower()
        else:  # pragma: no cover
            # no other types can be used in a layout.json display clause
            return

        # only column defined in install.json can be updated
        if column in self.ij.model.param_names:
            try:
                # value should be wrapped in single quotes to be properly parsed
                sql = f'''UPDATE {table_name} SET {column} = '{value}\''''
                cur = self.db_conn.cursor()
                cur.execute(sql)
            except sqlite3.OperationalError as e:  # pragma: no cover
                self.handle_error(f'SQL update failed - SQL: "{sql}", Error: "{e}"')

    def handle_error(self, err: str, halt: Optional[bool] = True):  # pragma: no cover
        """Print errors message and optionally exit.

        Args:
            err: The error message to print.
            halt: If True, the script will exit.
        """
        self.log.error(err)
        print(err)
        if halt:
            sys.exit(1)

    def get_action_input_names(self, action: str) -> List[str]:
        """Return the input names for the provided action."""
        return [i.name for i in self.get_action_inputs(action)]

    def get_action_inputs(self, action: str) -> List[InputModel]:
        """Return the inputs for the provided action."""
        return self.action_configurations.get(action, {}).get('inputs', [])

    def get_action_output_names(self, action: str) -> List[str]:
        """Return the output names for the provided action."""
        return [i.name for i in self.get_action_outputs(action)]

    def get_action_outputs(self, action: str) -> List[OutputsModel]:
        """Return the outputs for the provided action."""
        return self.action_configurations.get(action, {}).get('outputs', [])

    def get_input_applies_to_all(self, input_name: str) -> bool:
        """Return the outputs for the provided action."""
        for data in self.action_configurations.values():
            for d in data.get('inputs', []):
                if d.name == input_name:
                    break
            else:
                return False  # not found
        return True

    # TODO: [low] improve this logic
    def init_permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if self._input_permutations is None and self._output_permutations is None:
            self._input_permutations = []
            self._output_permutations = []

            # create db for permutations testing
            self.db_create_table(self._input_table, self.ij.model.param_names)
            self.db_insert_record(self._input_table, self.ij.model.param_names)

            # only gen permutations if none have been generated previously
            self._gen_permutations()

            # drop database
            self.db_drop_table(self._input_table)

    def input_dict(self, permutation_id: int) -> dict:
        """Return all input permutation names for provided permutation id.

        {'tc_action': 'Append', 'input_strings': None, 'append_chars': None}

        Args:
            permutation_id: The index of the permutation input array.

        Returns:
            dict: A dict with key / value for each input for the provided permutation id.
        """
        input_dict = {}
        if self.lj.has_layout:
            for permutation in self.input_permutations[permutation_id]:
                input_dict.setdefault(permutation.name, permutation.value)
        return input_dict

    @property
    def input_names(self) -> List[list]:
        """Return all input permutation names for current App.

        Returns:
            list: List of Lists of input names.
        """
        if self._input_names is None and self.lj.has_layout:
            self._input_names = []
            for permutation in self.input_permutations:
                self._input_names.append([p.name for p in permutation])
        return self._input_names

    @property
    def input_permutations(self) -> List[List[dict]]:
        """Return all input permutations for current App.

        self._input_permutations is an array of permutations arrays.
        [[<perm obj #1], [<perm obj #2]]

        Returns:
            list: List of Lists of valid input permutations.
        """
        if self._input_permutations is None and self.lj.has_layout:
            self.init_permutations()
        return self._input_permutations

    @property
    def output_permutations(self) -> List[List[dict]]:
        """Return all output permutations for current App.

        Returns:
            list: List of Lists of valid outputs permutations.
        """
        if self._output_permutations is None:
            self.init_permutations()
        return self._output_permutations

    def permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if not self.lj.has_layout:  # pragma: no cover
            print('Only Apps with a layout.json are supported.')
            sys.exit(1)

        if 'sqlite3' not in sys.modules:  # pragma: no cover
            print('The sqlite3 module needs to be built-in to Python for this feature.')
            sys.exit(1)

        # create db for permutations testing
        self.db_create_table(self._input_table, self.ij.model.param_names)
        self.db_insert_record(self._input_table, self.ij.model.param_names)

        # only gen permutations if none have been generated previously
        if not self._input_permutations and not self._output_permutations:
            self._input_permutations = self._input_permutations or []
            self._output_permutations = self._output_permutations or []
            self._gen_permutations()

        # output permutations
        self.write_permutations_file()

    def validate_input_variable(
        self, input_name: str, inputs: dict, display: Optional[str] = None
    ) -> bool:
        """Return True if the provided variables display where clause returns results.

        Args:
            input_name: The input variable name (e.g. tc_action).
            inputs: The current name/value dict.
            display: An optional display clause to validate. If not provided the
                display clause will be retrieved from the layout.json file.

        Returns:
            bool: True if the display value returns results.
        """
        # TODO: [low] re-evaluate this now that we only support layout.json Apps
        if not self.lj.has_layout or not inputs:  # pragma: no cover
            # always return true, even if current App doesn't have layouts
            return True

        table = f'temp_{random.randint(100,999)}'  # nosec
        self.db_create_table(table, self.ij.model.param_names)
        self.db_insert_record(table, self.ij.model.param_names)

        # APP-98 Added to cover the use case of interdependent variables in the layout.json.
        for name, param in self.ij.model.filter_params(_type='Boolean').items():
            self.db_update_record(table, name, param.default)

        for name, val in inputs.items():
            self.db_update_record(table, name, val)

        if display is None:
            lj_data = self.lj.model.get_param(input_name)
            if isinstance(lj_data, NoneModel):  # pragma: no cover
                # this shouldn't happen as all ij inputs must be in lj
                raise RuntimeError(f'The provided input {input_name} was not found in layout.json.')
            display = lj_data.display

        # check if provided variable meets display requirements
        valid = self.validate_layout_display(table, display)

        # cleanup temp table
        self.db_drop_table(table)

        return valid

    def validate_layout_display(self, table: str, display_condition: str) -> bool:
        """Check to see if the display condition passes.

        Args:
            table: The name of the DB table which hold the App data.
            display_condition: The "where" clause of the DB SQL statement.

        Returns:
            bool: True if the row count is greater than 0.
        """
        display = False
        if display_condition is None:
            display = True
        else:
            display_query = f'SELECT count(*) from {table} where {display_condition}'  # nosec
            try:
                cur = self.db_conn.cursor()
                cur.execute(display_query.replace('"', ''))
                rows = cur.fetchall()
                if rows[0][0] > 0:
                    display = True
            except sqlite3.Error as e:  # pragma: no cover
                print(f'"{display_query}" query returned an error: ({e}).')
                sys.exit(1)
        return display

    def write_permutations_file(self):
        """Print all valid permutations."""
        _permutations = []
        for index, permutations in enumerate(self.input_permutations):
            _permutations.append(
                {'index': index, 'args': [{'name': p.name, 'value': p.value} for p in permutations]}
            )

        with self.fqfn.open(mode='w') as fh:
            json.dump(_permutations, fh, indent=2, sort_keys=True)
        print('All permutations written to the "permutations.json" file.')



"""KeyValue Playbook Type"""
# standard library
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

# third-party
from pydantic import BaseModel, validator

# first-party
from tcex.input.field_types.binary import Binary
from tcex.input.field_types.exception import InvalidEmptyValue
from tcex.input.field_types.string import String
from tcex.input.field_types.tc_entity import TCEntity

if TYPE_CHECKING:  # pragma: no cover
    # third-party
    from pydantic.fields import ModelField


# pylint: disable=no-self-argument, no-self-use
class KeyValue(BaseModel):
    """Model for KeyValue Input."""

    key: str
    type: Optional[str]
    value: Union[
        List['KeyValue'],
        'KeyValue',
        List[TCEntity],
        TCEntity,
        List[String],
        String,
        List[Binary],
        Binary,
    ]

    @validator('key')
    def non_empty_string(cls, value: str, field: 'ModelField') -> Dict[str, Any]:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    class Config:
        """Model Config"""

        validate_assignment = True


KeyValue.update_forward_refs()



"""API Model"""
# pylint: disable=no-self-argument,no-self-use
# standard library
from typing import Optional

# third-party
from pydantic import BaseModel, Field, validator

# first-party
from tcex.app_config.install_json import InstallJson
from tcex.input.field_types.sensitive import Sensitive


class ApiModel(BaseModel):
    """API Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * TriggerService
    * WebhookTriggerService
    """

    api_default_org: Optional[str] = Field(
        None,
        description='The default ThreatConnect Org for the current API user.',
        inclusion_reason='runtimeLevel',
    )
    # alternate authentication credential when tc_token is not passed
    tc_api_access_id: Optional[str] = Field(
        None,
        description='A ThreatConnect API Access Id.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_api_path: str = Field(
        'https://api.threatconnect.com',
        description='The URL for the ThreatConnect API.',
        inclusion_reason='runtimeLevel',
    )
    # alternate authentication credential when tc_token is not passed
    tc_api_secret_key: Optional[Sensitive] = Field(
        None,
        description='A ThreatConnect API Secret Key.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_token: Optional[Sensitive] = Field(
        None,
        description='A ThreatConnect API token.',
        inclusion_reason='runtimeLevel',
    )
    tc_token_expires: Optional[int] = Field(
        None,
        description='The expiration timestamp in epoch for tc_token.',
        inclusion_reason='runtimeLevel',
    )
    tc_verify: Optional[bool] = Field(
        True,
        description='Flag to enable SSL validation for API requests.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )

    @validator('tc_token', always=True, pre=True)
    def one_set_of_credentials(cls, v, values):  # pylint: disable=E0213,R0201
        """Validate that one set of credentials is provided for the TC API."""
        _ij = InstallJson()

        # external Apps: require credentials and would not have an install.json file
        # organization (job) Apps: require credentials
        # playbook Apps: require credentials
        # service Apps: get token on createConfig message or during request
        if _ij.fqfn.is_file() is False or (
            _ij.model.is_playbook_app or _ij.model.is_organization_app
        ):
            if v is None and not all(
                [values.get('tc_api_access_id'), values.get('tc_api_secret_key')]
            ):
                raise ValueError(
                    'At least one set of ThreatConnect credentials must be provided '
                    '(tc_api_access_id/tc_api_secret key OR tc_token/tc_token_expires).'
                )
        return v



"""TcEx Generate Configurations CLI Command"""
# standard library
import json
import os

# third-party
from semantic_version import Version

# first-party
from tcex.app_config import AppSpecYml, InstallJson, JobJson, LayoutJson, TcexJson
from tcex.app_config.models import AppSpecYmlModel
from tcex.bin.bin_abc import BinABC


class SpecToolAppSpecYml(BinABC):
    """Generate App Config File"""

    def __init__(self):
        """Initialize class properties."""
        super().__init__()

        # properties
        self.filename = 'app_spec.yml'
        self.asy = AppSpecYml(logger=self.log)
        self.ij = InstallJson(logger=self.log)
        self.jj = JobJson(logger=self.log)
        self.lj = LayoutJson(logger=self.log)
        self.tj = TcexJson(logger=self.log)

    def _add_standard_fields(self, app_spec_yml_data: dict):
        """Add field that apply to ALL App types."""
        app_spec_yml_data.update(
            {
                'allowOnDemand': self.ij.model.allow_on_demand,
                'apiUserTokenParam': self.ij.model.api_user_token_param,
                'appId': str(self.ij.model.app_id),
                'category': self.ij.model.category,
                'deprecatesApps': self.ij.model.deprecates_apps,
                'displayName': self.ij.model.display_name,
                'features': self.ij.model.features,
                'labels': self.ij.model.labels,
                'languageVersion': self.ij.model.language_version,
                'listDelimiter': self.ij.model.list_delimiter,
                'minServerVersion': str(self.ij.model.min_server_version),
                'note': self.ij.model.note,
                'packageName': self.tj.model.package.app_name,
                'programLanguage': self.ij.model.program_language,
                'programMain': self.ij.model.program_main,
                'programVersion': str(self.ij.model.program_version),
                'runtimeLevel': self.ij.model.runtime_level,
            }
        )

    def _add_category(self, app_spec_yml_data: dict):
        """Add category."""
        _category = ''
        if any([self.ij.model.is_playbook_app, self.ij.model.is_trigger_app]):
            _category = self.ij.model.playbook.type or ''
        app_spec_yml_data['category'] = _category

    def _add_feeds(self, app_spec_yml_data: dict):
        """Add organization feeds section."""
        if not self.ij.model.is_feed_app:
            return

        feeds = []
        for feed in self.ij.model.feeds or []:
            if not os.path.isfile(feed.job_file):
                self.log.error(
                    f'feature=app-spec-yml, exception=failed-reading-file, filename={feed.job_file}'
                )
                continue
            with open(feed.job_file) as f:
                job = json.load(f)
            feed = feed.dict(by_alias=True)
            feed['job'] = job
            feeds.append(feed)
        app_spec_yml_data.setdefault('organization', {})
        app_spec_yml_data['organization']['feeds'] = feeds

    def _add_note_per_action(self, app_spec_yml_data: dict):
        """Add note per action."""
        _notes_per_action = []
        param = self.ij.model.get_param('tc_action')
        if param.valid_values:
            for action in param.valid_values:
                _notes_per_action.append({'action': action, 'note': ''})

            app_spec_yml_data['notesPerAction'] = _notes_per_action

    def _add_organization(self, app_spec_yml_data: dict):
        """Add asy.organization."""
        if self.ij.model.is_organization_app:
            app_spec_yml_data.setdefault('organization', {})
            if self.ij.model.publish_out_files:
                app_spec_yml_data['organization'][
                    'publishOutFiles'
                ] = self.ij.model.publish_out_files
            if self.ij.model.repeating_minutes:
                app_spec_yml_data['organization'][
                    'repeatingMinutes'
                ] = self.ij.model.repeating_minutes

    def _add_output_data(self, app_spec_yml_data: dict):
        """Add asy.outputData."""
        if any([self.ij.model.is_playbook_app, self.ij.model.is_trigger_app]):
            # build outputs based on display value
            _output_data_temp = {}
            if self.lj.has_layout:
                # layout based Apps could will have a display clause for each output
                for o in self.ij.model.playbook_outputs.values():
                    ljo = self.lj.model.get_output(o.name)
                    if ljo.display is not None and ljo.name is not None:
                        _output_data_temp.setdefault(o.display or '1', []).append(o.name)
                    else:
                        _output_data_temp.setdefault('1', []).append(o.name)
            else:
                for _, o in self.ij.model.playbook_outputs.items():
                    _output_data_temp.setdefault('1', []).append(o.name)

            _output_data = []
            for display, names in _output_data_temp.items():
                _output_variables = []
                for name in names:
                    if not self._is_advanced_request_output(name):
                        _output_variables.append(self.ij.model.get_output(name).dict(by_alias=True))

                _output_data.append({'display': display, 'outputVariables': _output_variables})

            app_spec_yml_data['outputData'] = _output_data

    def _add_output_prefix(self, app_spec_yml_data: dict):
        """Add asy.outputData."""
        if self.ij.model.is_playbook_app and self.ij.model.playbook.output_prefix:
            app_spec_yml_data['outputPrefix'] = self.ij.model.playbook.output_prefix

    def _add_playbook(self, app_spec_yml_data: dict):
        """Add asy.playbook."""
        if any([self.ij.model.is_playbook_app, self.ij.model.is_trigger_app]):
            app_spec_yml_data.setdefault('playbook', {})
            if self.ij.model.playbook.retry:
                app_spec_yml_data['playbook']['retry'] = self.ij.model.playbook.retry

    @staticmethod
    def _add_release_notes(app_spec_yml_data: dict):
        """Add release_notes."""
        app_spec_yml_data['releaseNotes'] = [
            {
                'notes': ['Initial Release'],
                'version': '1.0.0',
            }
        ]

    def _add_sections(self, app_spec_yml_data: dict):
        """Return params from ij and lj formatted for app_spec."""
        sections = []
        for section in self._current_data:
            _section_data = {'sectionName': section.get('title'), 'params': []}
            for p in section.get('parameters', []):
                if not self._is_advanced_request_input(p.get('name')):
                    param = self.ij.model.get_param(p.get('name')).dict(by_alias=True)
                    param['display'] = p.get('display')
                    _section_data['params'].append(param)
            sections.append(_section_data)
        app_spec_yml_data['sections'] = sections

    @property
    def _current_data(self):
        """Retrieve the appropriate data regardless of if its a layout based app."""
        if self.lj.has_layout:
            # handle layout based Apps
            _current_data = [i.dict(by_alias=True) for i in self.lj.model.inputs]

            # add hidden inputs from install.json (hidden inputs are not in layouts.json)
            _current_data.append(
                {
                    'parameters': [
                        p.dict(by_alias=True) for p in self.ij.model.params if p.hidden is True
                    ],
                    'title': 'Hidden Inputs',
                }
            )
        else:
            # handle non-layout based Apps
            _current_data = [
                {
                    'parameters': [p.dict(by_alias=True) for p in self.ij.model.params],
                    'title': 'Inputs',
                }
            ]
        return _current_data

    def _is_64_min_version(self) -> bool:
        """Return params from ij and lj formatted for app_spec."""

        for section in self._current_data:
            for p in section.get('parameters', []):
                param = self.ij.model.get_param(p.get('name')).dict(by_alias=True)
                if param['type'].lower() == 'editchoice':
                    return True
        return False

    @staticmethod
    def _is_advanced_request_input(name: str) -> bool:
        """Return true if input is an Advanced Request input."""
        return name in [
            'tc_adv_req_path',
            'tc_adv_req_http_method',
            'tc_adv_req_params',
            'tc_adv_req_exclude_null_params',
            'tc_adv_req_headers',
            'tc_adv_req_body',
            'tc_adv_req_urlencode_body',
            'tc_adv_req_fail_on_error',
        ]

    def _is_advanced_request_output(self, name: str) -> bool:
        """Return true if input is an Advanced Request input."""
        for pattern in [
            'request.content',
            'request.content.binary',
            'request.headers',
            'request.ok',
            'request.reason',
            'request.status_code',
            'request.url',
        ]:
            if f'{self.ij.model.playbook.output_prefix}.{pattern}' == name:
                return True
        return False

    def _add_min_tc_version(self, app_spec_yml_data: dict):
        """Add the correct min TC server version."""

        if self._is_64_min_version() and self.ij.model.min_server_version < Version('6.4.0'):
            app_spec_yml_data['minServerVersion'] = '6.4.0'

    def generate(self):
        """Generate the layout.json file data."""
        app_spec_yml_data = {}

        # add feeds
        self._add_feeds(app_spec_yml_data)

        # add release notes
        self._add_release_notes(app_spec_yml_data)

        # add standard fields
        self._add_standard_fields(app_spec_yml_data)

        # add category
        self._add_category(app_spec_yml_data)

        # add note per action
        self._add_note_per_action(app_spec_yml_data)

        # add organization (feed, jobs, etc)
        self._add_organization(app_spec_yml_data)

        # add playbook (retry)
        self._add_playbook(app_spec_yml_data)

        # add playbook output prefix
        self._add_output_prefix(app_spec_yml_data)

        # add sections
        self._add_sections(app_spec_yml_data)

        # add output data
        self._add_output_data(app_spec_yml_data)

        # parse and validate the data (use json so that json_encoder convert complex types)
        asy_data = json.loads(
            AppSpecYmlModel(**app_spec_yml_data).json(
                by_alias=True,
                exclude_defaults=True,
                exclude_none=True,
                exclude_unset=True,
                sort_keys=False,
            )
        )

        # force order of keys
        return self.asy.dict_to_yaml(self.asy.order_data(asy_data))

    @staticmethod
    def generate_schema():
        """Return the schema for the install.json file."""
        return AppSpecYmlModel.schema_json(indent=2, sort_keys=True)



"""Playbook delete."""
# standard library
import logging
from typing import Union

# first-party
from tcex.key_value_store import KeyValueApi, KeyValueRedis
from tcex.utils.utils import Utils

# get tcex logger
logger = logging.getLogger('tcex')


class PlaybookDelete:
    """Playbook Write ABC"""

    def __init__(
        self,
        context: str,
        key_value_store: Union[KeyValueApi, KeyValueRedis],
    ):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store

        # properties
        self.log = logger
        self.utils = Utils()

    def variable(self, key: str) -> str:
        """Delete method of CRUD operation for all data types.

        Only supported when using the Redis KV store.
        """
        data = None
        if key is not None:
            data = self.key_value_store.delete(self.context, key.strip())
        else:  # pragma: no cover
            self.log.warning('The key field was None.')
        return data



"""Logger module for TcEx Framework"""
# flake8: noqa
# first-party
# from tcex.logger.cache_handler import CacheHandler
# from tcex.logger.logger import Logger
# from tcex.logger.rotating_file_handler_custom import RotatingFileHandlerCustom
# from tcex.logger.trace_logger import TraceLogger



"""Model Definitions"""
# flake8: noqa
# first-party
from tcex.api.tc.ti_transform.model.transform_model import (
    AttributeTransformModel,
    DatetimeTransformModel,
    GroupTransformModel,
    IndicatorTransformModel,
    MetadataTransformModel,
    SecurityLabelTransformModel,
    TagTransformModel,
    TiTransformModel,
    TransformModel,
)



"""Model Definition"""
# standard library
from typing import Callable, List, Optional, Union

# third-party
from jmespath import compile as jmespath_compile
from pydantic import BaseModel, Extra, Field, root_validator, validator


# reusable validator
def _always_array(value: Union[list, str]) -> List[str]:
    """Return value as a list."""
    if not isinstance(value, list):
        value = [value]
    return value


# pylint: disable=no-self-argument, no-self-use
class TransformModel(BaseModel, extra=Extra.forbid):
    """Model Definition"""

    filter_map: Optional[dict] = Field(None, description='')
    kwargs: Optional[dict] = Field({}, description='')
    method: Optional[Callable] = Field(None, description='')
    for_each: Optional[Callable] = Field(None, description='')
    static_map: Optional[dict] = Field(None, description='')

    @validator('filter_map', 'static_map', pre=True)
    def _lower_map_keys(cls, v):
        """Validate static_map."""
        if isinstance(v, dict):
            v = {key.lower(): value for (key, value) in v.items()}
        return v

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _required_transform(cls, v: dict):
        """Validate either static_map or method is defined."""

        fields = [v.get('filter_map'), v.get('static_map'), v.get('method'), v.get('for_each')]

        if not any(fields):
            raise ValueError('Either map or method must be defined.')

        if fields.count(None) < 2:
            raise ValueError('Only one transform mechanism can be defined.')
        return v


class PathTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    default: Optional[str] = Field(None, description='')
    path: Optional[str] = Field(None, description='')

    @validator('path')
    def _validate_path(cls, v):
        """Validate path."""
        if v is not None:
            try:
                _ = jmespath_compile(v)
            except Exception:
                raise ValueError('A valid path must be provided.')
        return v

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _default_or_path(cls, values):
        """Validate either default or path is defined."""
        if not any([values.get('default'), values.get('path')]):
            raise ValueError('Either default or path or both must be defined.')
        return values


class MetadataTransformModel(PathTransformModel, extra=Extra.forbid):
    """."""

    transform: Optional[List[TransformModel]] = Field(None, description='')

    # validators
    _transform_array = validator('transform', allow_reuse=True, pre=True)(_always_array)


class ValueTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    value: Union[str, MetadataTransformModel]


class FileOccurrenceTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    file_name: Optional[Union[str, MetadataTransformModel]] = Field(None, description='')
    path: Optional[Union[str, MetadataTransformModel]] = Field(None, description='')
    date: Optional[Union[str, MetadataTransformModel]] = Field(None, description='')


class DatetimeTransformModel(PathTransformModel, extra=Extra.forbid):
    """."""


class AssociatedGroupTransform(ValueTransformModel, extra=Extra.forbid):
    """."""


class AttributeTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""

    displayed: Union[bool, MetadataTransformModel] = Field(False, description='')
    source: Optional[Union[str, MetadataTransformModel]] = Field(None, description='')
    type: Union[str, MetadataTransformModel] = Field(..., description='')


class SecurityLabelTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""

    color: Optional[MetadataTransformModel] = Field(None, description='')
    description: Optional[MetadataTransformModel] = Field(None, description='')


class TagTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""


class TiTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    applies: Optional[Callable] = Field(None, description='')
    associated_groups: Optional[List[AssociatedGroupTransform]] = Field(None, description='')
    attributes: Optional[List[AttributeTransformModel]] = Field(None, description='')
    date_added: Optional[DatetimeTransformModel] = Field(None, description='')
    last_modified: Optional[DatetimeTransformModel] = Field(None, description='')
    security_labels: Optional[List[SecurityLabelTransformModel]] = Field(None, description='')
    tags: Optional[List[TagTransformModel]] = Field(None, description='')
    type: MetadataTransformModel = Field(..., description='')
    xid: Optional[MetadataTransformModel] = Field(description='')


class GroupTransformModel(TiTransformModel, extra=Extra.forbid):
    """."""

    name: MetadataTransformModel = Field(..., description='')
    # campaign
    first_seen: Optional[DatetimeTransformModel] = Field(None, description='')
    # document
    malware: Optional[MetadataTransformModel] = Field(None, description='')
    password: Optional[MetadataTransformModel] = Field(None, description='')
    # email
    from_addr: Optional[MetadataTransformModel] = Field(None, description='')
    score: Optional[MetadataTransformModel] = Field(None, description='')
    to_addr: Optional[MetadataTransformModel] = Field(None, description='')
    # event, incident
    event_date: Optional[DatetimeTransformModel] = Field(None, description='')
    status: Optional[MetadataTransformModel] = Field(None, description='')
    # report
    publish_date: Optional[DatetimeTransformModel] = Field(None, description='')
    # signature
    file_type: Optional[MetadataTransformModel] = Field(None, description='')
    file_text: Optional[MetadataTransformModel] = Field(None, description='')
    # document, signature
    file_name: Optional[MetadataTransformModel] = Field(None, description='')


# pylint: disable=no-self-argument,no-self-use
class IndicatorTransformModel(TiTransformModel, extra=Extra.forbid):
    """."""

    confidence: Optional[MetadataTransformModel] = Field(None, description='')
    rating: Optional[MetadataTransformModel] = Field(None, description='')
    # summary: Optional[MetadataTransformModel] = Field(None, description='')
    value1: Optional[MetadataTransformModel] = Field(None, description='')
    value2: Optional[MetadataTransformModel] = Field(None, description='')
    value3: Optional[MetadataTransformModel] = Field(None, description='')
    # file
    file_occurrences: Optional[List[FileOccurrenceTransformModel]] = Field(None, description='')
    size: Optional[MetadataTransformModel] = Field(None, description='')
    # host
    dns_active: Optional[MetadataTransformModel] = Field(None, description='')
    whois_active: Optional[MetadataTransformModel] = Field(None, description='')

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _one_indicator_value(cls, values):
        """Validate that one set of credentials is provided for the TC API."""

        if not any(
            [
                values.get('value1'),
                values.get('value2'),
                values.get('value3'),
            ]
        ):
            raise ValueError(
                'An indicator transform must be defined (e.g., summary, value1, value2, value3).'
            )
        return values



"""TcEx Generate Configurations CLI Command"""
# standard library
from typing import TYPE_CHECKING, List, Optional

# first-party
from tcex.app_config.layout_json import LayoutJson
from tcex.app_config.permutation import Permutation
from tcex.bin.bin_abc import BinABC

if TYPE_CHECKING:
    # first-party
    from tcex.app_config import AppSpecYml
    from tcex.app_config.models.app_spec_yml_model import SectionsModel
    from tcex.app_config.models.install_json_model import ParamsModel


class SpecToolReadmeMd(BinABC):
    """Generate App Config File"""

    def __init__(self, asy: 'AppSpecYml'):
        """Initialize class properties."""
        super().__init__()
        self.asy = asy

        # properties
        self.i1 = ''
        self.filename = 'README.md'
        self.lj = LayoutJson(logger=self.log)
        self.permutations = Permutation(self.log)

    def _add_actions_title(self, readme_md: List[str]):
        """Add title for action section."""
        readme_md.append(self._markdown_header1('Actions'))
        readme_md.append('')
        readme_md.append(self._markdown_horizontal_rule)
        readme_md.append('')

    def _add_actions_sub_title(self, readme_md: List[str], action: str):
        """Add title for sub action section."""
        readme_md.append(self._markdown_header2(action))
        readme_md.append('')
        npa = self.asy.model.get_note_per_action(action).note
        if npa is not None:
            readme_md.append(self.asy.model.get_note_per_action(action).note)
            readme_md.append('')

    def _add_labels(self, readme_md: List[str]):
        """Add labels data to readme.md."""
        if self.asy.model.labels:
            readme_md.append(self._markdown_header1('Labels'))
            readme_md.append('')
            _labels = ', '.join(sorted(self.asy.model.labels))
            readme_md.append(f'-   {_labels}')

    def _add_description(self, readme_md: List[str]):
        """Add top level description/note data to readme.md."""
        if self.asy.model.note:
            readme_md.append(self._markdown_header1('Description'))
            readme_md.append('')
            readme_md.append(self.asy.model.note)
            readme_md.append('')
            if self.asy.model.note_per_action:
                readme_md.append('\n\n'.join(self.asy.model.note_per_action_formatted))
                readme_md.append('')

        if self.asy.model.service_details:
            readme_md.append('')
            readme_md.append(self.asy.model.service_details)

    def _add_inputs_title(self, readme_md: List[str]):
        """Add title for input section."""
        readme_md.append(f'{self._markdown_header_input} Inputs')
        readme_md.append('')

    def _add_service_config_title(self, readme_md: List[str]):
        """Add title for service configuration section."""
        readme_md.append(self._markdown_header1('Service Configuration'))
        readme_md.append('')

    def _add_param(self, readme_md: List[str], param: 'ParamsModel'):
        """Add params data to readme.md.

        **API Key** *(String)*
        _**Duration**_ *(String, Optional)*
        """
        label = self._markdown_bold(param.label)
        type_data = f'{param.type}'

        if param.required is False:
            # change the format of the label name to italics if it is optional
            if param.type.lower() not in ['boolean']:
                label = self._markdown_italic_nested(label)
                type_data += ', Optional'

        if param.default is not None:
            # following current format where boolean values are shown as
            # selected/unselected and others true/false
            if param.type.lower() == 'boolean':
                default_value = 'Selected' if param.default is True else 'Unselected'
            else:
                default_value = param.default
            type_data += f''', Default: {str(default_value).replace('|', ', ')}'''

        type_data_formatted = self._markdown_italic(f'({type_data})')
        readme_md.append(f'{self.i1}{label} {type_data_formatted}')
        readme_md.append('')

    def _add_params(
        self, readme_md: List[str], section: 'SectionsModel', action: Optional[str] = None
    ):
        # add params
        for param in section.params:
            if any([param.disabled, param.hidden]):
                continue

            # don't add tc_action param since it's the top level action
            if param.name == 'tc_action':
                continue

            if action is not None:
                if param.name not in self.permutations.get_action_input_names(action):
                    continue

            # add param data
            self._add_param(readme_md, param)

            # add param note data
            self._add_param_note(readme_md, param)

            # add param playbook data types data
            self._add_param_pb_data_type(readme_md, param)

            # add param valid_values data
            self._add_param_valid_values(readme_md, param)

    def _add_param_note(self, readme_md: List[str], param: 'ParamsModel'):
        """Add note data to readme.md."""
        if param.note:
            readme_md.append(f'{self.i1}{param.note}')
            readme_md.append('')

    def _add_param_pb_data_type(self, readme_md: List[str], param: 'ParamsModel'):
        """Add playbook data types values data to readme.md."""
        # matching current format where single 'String' is not displayed
        if param.playbook_data_type and param.playbook_data_type != ['String']:
            _pdt = ', '.join(param.playbook_data_type)
            readme_md.append(f'''{self.i1}> {self._markdown_bold('Allows:')} {_pdt}''')
            readme_md.append('')

    def _add_param_valid_values(self, readme_md: List[str], param: 'ParamsModel'):
        """Add valid values data to readme.md."""
        # matching current format where TEXT and KEYCHAIN were excluded.
        valid_values = [p for p in param.valid_values if not p.startswith('${')]
        if valid_values:
            _valid_values = ', '.join(valid_values)
            readme_md.append(
                f'''{self.i1}> {self._markdown_bold('Valid Values:')} {_valid_values}'''
            )
            readme_md.append('')

    def _add_outputs(self, readme_md: List[str], action: str = None):
        """Add output data to readme.md."""
        if self.asy.model.output_variables:
            readme_md.append(f'{self._markdown_header_output} Outputs')
            readme_md.append('')

            outputs = self.ij.model.playbook.output_variables
            outputs = sorted(list(set(outputs)), key=lambda x: x.name)
            if action:
                outputs = self.permutations.get_action_outputs(action)

            for output in outputs:
                output_type = self._markdown_italic(f'({output.type})')
                readme_md.append(f'{self.i1}-   {output.name} {output_type}')

            readme_md.append('')

    def _add_section_title(self, readme_md: List[str], section: 'SectionsModel'):
        """Add title for input section."""
        readme_md.append(self._markdown_header3(self._markdown_italic(section.section_name)))
        readme_md.append('')

    def _add_params_for_playbook_action_app(self, readme_md: List[str], actions: List[str]):
        """Add inputs for playbook action app."""
        # add title for actions section
        self._add_actions_title(readme_md)

        for action in actions:

            # add title for action sub section
            self._add_actions_sub_title(readme_md, action)

            # add inputs and sections
            self._add_inputs_title(readme_md)

            for section in self.asy.model.sections:
                # don't show the section if it has no params
                if self._has_section_params(section, action) is False:
                    continue

                # add section title
                self._add_section_title(readme_md, section)

                # add params
                self._add_params(readme_md, section, action)

            # add output data
            self._add_outputs(readme_md, action)

            # add horizontal rule
            readme_md.append(self._markdown_horizontal_rule)
            readme_md.append('')

    def _add_params_for_playbook_std_app(self, readme_md: List[str]):
        """Add inputs for playbook standard app."""
        self._add_inputs_title(readme_md)

        for section in self.asy.model.sections:
            # don't show the section if it has no params
            valid_section = False
            for sp in section.params:
                if sp.disabled is False and sp.hidden is False:
                    valid_section = True

            if valid_section is False:
                continue

            # add section title
            self._add_section_title(readme_md, section)

            self._add_params(readme_md, section)

            # add output data
            self._add_outputs(readme_md)

    def _add_params_for_non_playbook_apps(self, readme_md: List[str]):
        """Add inputs for non playbook app."""
        service_config = []
        non_service_config = []

        # Separate Params into service configuration params and other parameters
        for param in self.asy.model.params:
            if param.disabled is True or param.hidden is True:
                continue
            if param.service_config is True:
                service_config.append(param)
            else:
                non_service_config.append(param)

        # Add service configuration params to ReadMe file.
        if service_config:
            self._add_service_config_title(readme_md)

            for param in service_config:
                # add param data
                self._add_param(readme_md, param)

                # add param note data
                self._add_param_note(readme_md, param)

                # add param valid_values data
                self._add_param_valid_values(readme_md, param)

        # add inputs and sections
        self._add_inputs_title(readme_md)

        for param in non_service_config:
            # add param data
            self._add_param(readme_md, param)

            # add param note data
            self._add_param_note(readme_md, param)

            # add param valid_values data
            self._add_param_valid_values(readme_md, param)

        # add output data
        self._add_outputs(readme_md)

    def _has_section_params(self, section: 'SectionsModel', action: str) -> bool:
        """Return True if the provided section has params."""
        if [
            sp
            for sp in section.params
            if not any([sp.disabled, sp.hidden, sp.service_config])
            and sp.name != 'tc_action'
            and self._valid_param_for_action(sp, action) is True
        ]:
            return True
        return False

    @staticmethod
    def _markdown_bold(string: str):
        """Return bold string."""
        return f'**{string}**'

    @staticmethod
    def _markdown_bullet_list_item(string: str):
        """Return italic string."""
        return f'-   {string}'

    @property
    def _markdown_header_input(self):
        """Return the appropriate header level for input section."""
        if self.lj.has_layout:
            return '###'
        return '#'

    @property
    def _markdown_header_output(self):
        """Return the appropriate header level for output section."""
        if self.lj.has_layout:
            return '###'
        return '#'

    @staticmethod
    def _markdown_header1(string: str):
        """Return header string."""
        return f'# {string}'

    @staticmethod
    def _markdown_header2(string: str):
        """Return header string."""
        return f'## {string}'

    @staticmethod
    def _markdown_header3(string: str):
        """Return header string."""
        return f'### {string}'

    @property
    def _markdown_horizontal_rule(self):
        """Return markdown horizontal rule."""
        return '---'

    @staticmethod
    def _markdown_italic(string: str):
        """Return italic string."""
        return f'*{string}*'

    @staticmethod
    def _markdown_italic_nested(string: str):
        """Return italic string."""
        return f'_{string}_'

    def _valid_param_for_action(self, param: 'ParamsModel', action: str) -> bool:
        """Return True if param is valid for action."""
        for input_param in self.permutations.get_action_inputs(action):
            if input_param.name == param.name:
                return True
        return False

    def generate(self) -> List[str]:
        """Generate the README.md file data."""
        readme_md = []

        # add App Name
        readme_md.append(self._markdown_header1(self.asy.model.display_name))
        readme_md.append('')

        # add release notes
        readme_md.extend(self.asy.model.release_notes_formatted)

        # add category
        if self.asy.model.category:
            readme_md.append(self._markdown_header1('Category'))
            readme_md.append('')
            readme_md.append(self._markdown_bullet_list_item(self.asy.model.category))
            readme_md.append('')

        # add description
        self._add_description(readme_md)

        # add inputs
        if self.asy.model.is_playbook_app:
            actions = self.ij.model.get_param('tc_action').valid_values or []
            if actions:
                # add inputs for action based sections
                self._add_params_for_playbook_action_app(readme_md, actions)
            else:
                # add inputs for non action based sections
                self._add_params_for_playbook_std_app(readme_md)

        elif self.asy.model.is_organization_app or self.asy.model.is_trigger_app:
            self._add_params_for_non_playbook_apps(readme_md)

        # add labels
        self._add_labels(readme_md)

        # add end of file newline
        readme_md.append('')
        return readme_md



"""Case / Cases Object"""
# standard library
from typing import TYPE_CHECKING, Iterator, Union

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.case_attributes.case_attribute_model import CaseAttributeModel
from tcex.api.tc.v3.cases.case_filter import CaseFilter
from tcex.api.tc.v3.cases.case_model import CaseModel, CasesModel
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.tags.tag_model import TagModel
from tcex.api.tc.v3.tasks.task_model import TaskModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact
    from tcex.api.tc.v3.case_attributes.case_attribute import CaseAttribute
    from tcex.api.tc.v3.groups.group import Group
    from tcex.api.tc.v3.indicators.indicator import Indicator
    from tcex.api.tc.v3.notes.note import Note
    from tcex.api.tc.v3.tags.tag import Tag
    from tcex.api.tc.v3.tasks.task import Task


class Cases(ObjectCollectionABC):
    """Cases Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = CasesModel(**kwargs)
        self.type_ = 'cases'

    def __iter__(self) -> 'Case':
        """Iterate over CM objects."""
        return self.iterate(base_class=Case)

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.CASES.value

    @property
    def filter(self) -> 'CaseFilter':
        """Return the type specific filter object."""
        return CaseFilter(self.tql)


class Case(ObjectABC):
    """Cases Object.

    Args:
        artifacts (Artifacts, kwargs): A list of Artifacts corresponding to the Case.
        assignee (Assignee, kwargs): The user or group Assignee object for the Case.
        associated_cases (Cases, kwargs): A list of Cases associated with this Case.
        associated_groups (Groups, kwargs): A list of Groups associated with this Case.
        associated_indicators (Indicators, kwargs): A list of Indicators associated with this Case.
        attributes (CaseAttributes, kwargs): A list of Attributes corresponding to the Case.
        case_close_time (str, kwargs): The date and time that the Case was closed.
        case_detection_time (str, kwargs): The date and time that ends the user initiated Case
            duration.
        case_occurrence_time (str, kwargs): The date and time that starts the user initiated Case
            duration.
        case_open_time (str, kwargs): The date and time that the Case was first opened.
        description (str, kwargs): The description of the Case.
        name (str, kwargs): The name of the Case.
        notes (Notes, kwargs): A list of Notes corresponding to the Case.
        resolution (str, kwargs): The Case resolution.
        severity (str, kwargs): The Case severity.
        status (str, kwargs): The Case status.
        tags (Tags, kwargs): A list of Tags corresponding to the Case (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        tasks (Tasks, kwargs): A list of Tasks corresponding to the Case.
        user_access (Users, kwargs): A list of Users that, when defined, are the only ones allowed
            to view or edit the Case.
        workflow_events (WorkflowEvents, kwargs): A list of workflowEvents (timeline) corresponding
            to the Case.
        workflow_template (WorkflowTemplate, kwargs): The Template that the Case is populated by.
        xid (str, kwargs): The **xid** for the Case.
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model = CaseModel(**kwargs)
        self._nested_field_name = 'cases'
        self._nested_filter = 'has_case'
        self.type_ = 'Case'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.CASES.value

    @property
    def model(self) -> 'CaseModel':
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: Union['CaseModel', dict]):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    @property
    def artifacts(self) -> Iterator['Artifact']:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)

    @property
    def associated_cases(self) -> Iterator['Case']:
        """Yield Case from Cases."""
        # Ensure the current item is not returned as a association
        for case in self._iterate_over_sublist(Cases):
            if case.model.id == self.model.id:
                continue
            yield case

    @property
    def associated_groups(self) -> Iterator['Group']:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)

    @property
    def associated_indicators(self) -> Iterator['Indicator']:
        """Yield Indicator from Indicators."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator import Indicators

        yield from self._iterate_over_sublist(Indicators)

    @property
    def attributes(self) -> Iterator['CaseAttribute']:
        """Yield Attribute from Attributes."""
        # first-party
        from tcex.api.tc.v3.case_attributes.case_attribute import CaseAttributes

        yield from self._iterate_over_sublist(CaseAttributes)

    @property
    def notes(self) -> Iterator['Note']:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)

    @property
    def tags(self) -> Iterator['Tag']:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)

    @property
    def tasks(self) -> Iterator['Task']:
        """Yield Task from Tasks."""
        # first-party
        from tcex.api.tc.v3.tasks.task import Tasks

        yield from self._iterate_over_sublist(Tasks)

    def stage_artifact(self, data: Union[dict, 'ObjectABC', 'ArtifactModel']):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            raise RuntimeError('Invalid type passed in to stage_artifact')
        data._staged = True
        self.model.artifacts.data.append(data)

    # pylint: disable=redefined-builtin
    def stage_assignee(self, type: str, data: Union[dict, 'ObjectABC', 'ArtifactModel']):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif type.lower() == 'user' and isinstance(data, dict):
            data = UserModel(**data)
        elif type.lower() == 'group' and isinstance(data, dict):
            data = UserGroupModel(**data)

        if not isinstance(data, (UserModel, UserGroupModel)):
            raise RuntimeError('Invalid type passed in to stage_assignee')
        data._staged = True
        self.model.assignee._staged = True
        self.model.assignee.type = type
        self.model.assignee.data = data

    def stage_associated_case(self, data: Union[dict, 'ObjectABC', 'CaseModel']):
        """Stage case on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = CaseModel(**data)

        if not isinstance(data, CaseModel):
            raise RuntimeError('Invalid type passed in to stage_associated_case')
        data._staged = True
        self.model.associated_cases.data.append(data)

    def stage_associated_group(self, data: Union[dict, 'ObjectABC', 'GroupModel']):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            raise RuntimeError('Invalid type passed in to stage_associated_group')
        data._staged = True
        self.model.associated_groups.data.append(data)

    def stage_associated_indicator(self, data: Union[dict, 'ObjectABC', 'IndicatorModel']):
        """Stage indicator on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = IndicatorModel(**data)

        if not isinstance(data, IndicatorModel):
            raise RuntimeError('Invalid type passed in to stage_associated_indicator')
        data._staged = True
        self.model.associated_indicators.data.append(data)

    def stage_attribute(self, data: Union[dict, 'ObjectABC', 'CaseAttributeModel']):
        """Stage attribute on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = CaseAttributeModel(**data)

        if not isinstance(data, CaseAttributeModel):
            raise RuntimeError('Invalid type passed in to stage_attribute')
        data._staged = True
        self.model.attributes.data.append(data)

    def stage_note(self, data: Union[dict, 'ObjectABC', 'NoteModel']):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            raise RuntimeError('Invalid type passed in to stage_note')
        data._staged = True
        self.model.notes.data.append(data)

    def stage_tag(self, data: Union[dict, 'ObjectABC', 'TagModel']):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            raise RuntimeError('Invalid type passed in to stage_tag')
        data._staged = True
        self.model.tags.data.append(data)

    def stage_task(self, data: Union[dict, 'ObjectABC', 'TaskModel']):
        """Stage task on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = TaskModel(**data)

        if not isinstance(data, TaskModel):
            raise RuntimeError('Invalid type passed in to stage_task')
        data._staged = True
        self.model.tasks.data.append(data)

    def stage_user_access(self, data: Union[dict, 'ObjectABC', 'UserModel']):
        """Stage user on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = UserModel(**data)

        if not isinstance(data, UserModel):
            raise RuntimeError('Invalid type passed in to stage_user_access')
        data._staged = True
        self.model.user_access.data.append(data)



"""Model Map"""

# first-party
from tcex.input.models.advanced_request_model import _AdvancedRequestModel
from tcex.input.models.aot_execution_enabled_model import AotExecutionEnabledModel
from tcex.input.models.api_model import ApiModel
from tcex.input.models.batch_model import BatchModel
from tcex.input.models.cal_settings_model import CalSettingsModel
from tcex.input.models.logging_model import LoggingModel
from tcex.input.models.organization_model import OrganizationModel
from tcex.input.models.path_model import PathModel
from tcex.input.models.playbook_common_model import PlaybookCommonModel
from tcex.input.models.playbook_model import PlaybookModel
from tcex.input.models.proxy_model import ProxyModel
from tcex.input.models.service_model import ServiceModel
from tcex.input.models.smtp_settings_model import SmtpSettingsModel

# define feature to model map
feature_map = {
    'aotExecutionEnabled': [AotExecutionEnabledModel],
    'appBuilderCompliant': [],
    # 'advancedRequest': [],
    'CALSettings': [CalSettingsModel],
    'fileParams': [],
    'layoutEnabledApp': [],
    'secureParams': [],
    'smtpSettings': [SmtpSettingsModel],
    # features for TC Playbook loop prevention
    'CreatesGroup': [],
    'CreatesIndicator': [],
    'CreatesSecurityLabel': [],
    'CreatesTag': [],
    'DeletesGroup': [],
    'DeletesIndicator': [],
    'DeletesSecurityLabel': [],
    'DeletesTag': [],
}

# define runtime level to model map
runtime_level_map = {
    'apiservice': [
        ApiModel,
        BatchModel,
        LoggingModel,
        PathModel,
        PlaybookCommonModel,
        ProxyModel,
        ServiceModel,
    ],
    'feedapiservice': [
        ApiModel,
        BatchModel,
        LoggingModel,
        PathModel,
        PlaybookCommonModel,
        ProxyModel,
        ServiceModel,
    ],
    'external': [
        ApiModel,
        BatchModel,
        LoggingModel,
        ProxyModel,
    ],
    'organization': [
        ApiModel,
        BatchModel,
        LoggingModel,
        OrganizationModel,
        PathModel,
        ProxyModel,
    ],
    'playbook': [
        ApiModel,
        BatchModel,
        LoggingModel,
        PathModel,
        PlaybookCommonModel,
        PlaybookModel,
        ProxyModel,
        # ServiceModel,
    ],
    'triggerservice': [
        ApiModel,
        BatchModel,
        LoggingModel,
        PathModel,
        PlaybookCommonModel,
        ProxyModel,
        ServiceModel,
    ],
    'webhooktriggerservice': [
        ApiModel,
        BatchModel,
        LoggingModel,
        PathModel,
        PlaybookCommonModel,
        ProxyModel,
        ServiceModel,
    ],
}

tc_action_map = {
    'Advanced Request': [_AdvancedRequestModel],
}



"""ThreatConnect Assignee Module"""
# pylint: disable=no-member,no-self-argument,no-self-use,wrong-import-position
# standard library
from typing import Optional, Union

# third-party
from pydantic import Field, validator

# first-party
from tcex.api.tc.v3.security.assignee_user_group_model import AssigneeUserGroupModel
from tcex.api.tc.v3.security.assignee_user_model import AssigneeUserModel
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.utils import Utils


class AssigneeModel(
    V3ModelABC,
    title='User Data Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Assignee Model"""

    type: Optional[str] = Field(
        None,
        description='The **Type** for the Assignee.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='type',
    )

    data: Optional[Union[AssigneeUserModel, AssigneeUserGroupModel]] = Field(
        None,
        description='The **Data** for the Assignee.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='data',
    )

    @validator('type', always=True)
    def _validate_type(cls, v):
        if not v:
            return v

        if v.lower() == 'user':
            return 'User'

        if v.lower() in ['group', 'user_group', 'usergroup']:
            return 'Group'

        raise ValueError('Value must be either `User` or `Group`.')



"""IndicatorAttribute / IndicatorAttributes Object"""
# standard library
from typing import TYPE_CHECKING, Iterator, Union

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.indicator_attributes.indicator_attribute_filter import IndicatorAttributeFilter
from tcex.api.tc.v3.indicator_attributes.indicator_attribute_model import (
    IndicatorAttributeModel,
    IndicatorAttributesModel,
)
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.security_labels.security_label import SecurityLabel


class IndicatorAttributes(ObjectCollectionABC):
    """IndicatorAttributes Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = IndicatorAttributesModel(**kwargs)
        self.type_ = 'indicator_attributes'

    def __iter__(self) -> 'IndicatorAttribute':
        """Iterate over CM objects."""
        return self.iterate(base_class=IndicatorAttribute)

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INDICATOR_ATTRIBUTES.value

    @property
    def filter(self) -> 'IndicatorAttributeFilter':
        """Return the type specific filter object."""
        return IndicatorAttributeFilter(self.tql)


class IndicatorAttribute(ObjectABC):
    """IndicatorAttributes Object.

    Args:
        default (bool, kwargs): A flag indicating that this is the default attribute of its type
            within the object. Only applies to certain attribute and data types.
        indicator_id (int, kwargs): Indicator associated with attribute.
        pinned (bool, kwargs): A flag indicating that the attribute has been noted for importance.
        security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
            Intel item (NOTE: Setting this parameter will replace any existing tag(s) with
            the one(s) specified).
        source (str, kwargs): The attribute source.
        type (str, kwargs): The attribute type.
        value (str, kwargs): The attribute value.
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model = IndicatorAttributeModel(**kwargs)
        self._nested_field_name = 'attributes'
        self._nested_filter = 'has_indicator_attribute'
        self.type_ = 'Indicator Attribute'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INDICATOR_ATTRIBUTES.value

    @property
    def model(self) -> 'IndicatorAttributeModel':
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: Union['IndicatorAttributeModel', dict]):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def security_labels(self) -> Iterator['SecurityLabel']:
        """Yield SecurityLabel from SecurityLabels."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label import SecurityLabels

        yield from self._iterate_over_sublist(SecurityLabels)

    def stage_security_label(self, data: Union[dict, 'ObjectABC', 'SecurityLabelModel']):
        """Stage security_label on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = SecurityLabelModel(**data)

        if not isinstance(data, SecurityLabelModel):
            raise RuntimeError('Invalid type passed in to stage_security_label')
        data._staged = True
        self.model.security_labels.data.append(data)



"""None Model"""


class Singleton(type):
    """A singleton Metaclass"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Evoke call method."""
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NoneModel(metaclass=Singleton):
    """A dummy model that return None for all attribute requests."""

    def __getattribute__(self, name: str):
        """Return None for any attribute request."""
        return None



"""TcEx Generate Configurations CLI Command"""
# standard library
import os
import re
from typing import TYPE_CHECKING, List, Optional, Tuple

# first-party
import tcex.input.field_types as FieldTypes  # noqa: N812
from tcex.app_config.permutation import Permutation
from tcex.backports import cached_property
from tcex.bin.bin_abc import BinABC
from tcex.bin.spec_tool_app_input_static import SpecToolAppInputStatic
from tcex.pleb.none_model import NoneModel

if TYPE_CHECKING:
    # first-party
    from tcex.app_config.models.install_json_model import ParamsModel


class SpecToolAppInput(BinABC):
    """Generate App Config File"""

    def __init__(self):
        """Initialize class properties."""
        super().__init__()

        # properties
        self._app_inputs_data = None
        self.class_model_map = {}
        self.field_type_modules = set()
        self.filename = 'app_inputs.py'
        self.input_static = SpecToolAppInputStatic()
        self.typing_modules = set()
        self.pydantic_modules = {'BaseModel'}
        self.permutations = Permutation(self.log)
        self.report_mismatch = []

    def _add_action_classes(self):
        """Add actions to the App."""
        for action in self._tc_actions:
            class_name = self._gen_tc_action_class_name(action)
            self.class_model_map[action] = class_name
            self._app_inputs_data[class_name] = {}

    def _add_input_to_action_class(
        self, applies_to_all: bool, param_data: 'ParamsModel', tc_action: Optional[str] = None
    ):
        """Add input data to Action class."""
        tc_action_class = 'AppBaseModel'
        if applies_to_all is False:
            tc_action_class = self._gen_tc_action_class_name(tc_action)
        self.app_inputs_data[tc_action_class][param_data.name] = param_data

    def _add_typing_import_module(self, type_: str):
        """Add the appropriate import module for typing."""
        if 'List[' in type_:
            self.typing_modules.add('List')

        if 'Optional[' in type_:
            self.typing_modules.add('Optional')

        if 'Union[' in type_:
            self.typing_modules.add('Union')

    def class_comment(self, model_class: str) -> str:
        """Return the appropriate comment for the provided class."""
        _class_comment = f'{self.i1}"""Action Model"""'
        _comment_map = {
            'AppBaseModel': self.input_static.app_base_model_class_comment,
            'ServiceConfigModel': self.input_static.service_config_model_class_comment,
            'TriggerConfigModel': self.input_static.trigger_config_model_class_comment,
        }
        return _comment_map.get(model_class, _class_comment)

    @property
    def _code_app_inputs_data(self):
        """Return app_inputs.py input data."""
        _code = []
        # sorting of the inputs can only be done at this point
        for class_name, class_inputs in self.app_inputs_data.items():
            # determine the base class for the current action class
            base_class = 'AppBaseModel'
            class_comment = self.class_comment(class_name)
            if class_name in ['AppBaseModel', 'ServiceConfigModel']:
                base_class = 'BaseModel'
            elif class_name == 'TriggerConfigModel':
                base_class = 'CreateConfigModel'

            _code.extend(
                [
                    f'class {class_name}({base_class}):',
                    f'{class_comment}',
                    '',
                ]
            )
            _always_array = []
            _entity_input = []
            for input_name, input_data in sorted(class_inputs.items()):
                # skip special inputs
                if input_name in ['tc_log_level']:
                    continue

                # add comment
                _comment = self._code_app_inputs_data_comments(input_data)
                if _comment is not None:
                    _code.append(_comment)

                # add finalized input code line
                _type_data = f'{self.i1}{input_name}: {self._gen_type(class_name, input_data)}'
                _code.append(_type_data)

                # check if validator are applicable to this input
                if self._validator_always_array_check(input_data) is True:
                    _always_array.append(f'\'{input_name}\'')

                if self._validator_entity_input_check(input_data) is True:
                    _entity_input.append(f'\'{input_name}\'')

            # validator - always_array
            if _always_array:
                self.field_type_modules.add('always_array')
                self.pydantic_modules.add('validator')
                _code.extend(self._validator_always_array(_always_array))

            # validator - entity_input
            if _entity_input:
                self.field_type_modules.add('entity_input')
                self.pydantic_modules.add('validator')
                _code.extend(self._validator_entity_input(_entity_input))

        # append 2 blank lines between classes
        _code.append('')
        _code.append('')
        return _code

    def _code_app_inputs_data_comments(self, input_data: 'ParamsModel') -> str:
        """Return comments for a single input."""
        # append comment for playbookDataTypes
        comments = []
        if input_data.playbook_data_type:
            playbook_data_type_str = '|'.join(input_data.playbook_data_type)
            comments.append(f'pbd: {playbook_data_type_str}')
        # append comment for validValues
        if input_data.valid_values:
            valid_values_str = '|'.join(input_data.valid_values)
            comments.append(f'vv: {valid_values_str}')

        if comments:
            comment = ', '.join(comments)
            comment_wrapped = self.utils.wrap_string(comment, [' ', '|'], 80).split('\n')
            return '\n'.join([f'{self.i1}# {c}' for c in comment_wrapped])
        return None

    def _extract_type_from_definition(self, input_name: str, type_definition: str) -> str:
        """Extract the type from the type definition.

        string_allow_multiple: Union[String, List[String]] -> Union[String, List[String]]
        string_intel_type: Optional[String] -> Optional[str]
        """
        input_extract_pattern = (
            # match beginning white space on line
            r'(?:^\s+)?'
            # match input name
            rf'(?:{input_name}:\s)'
            # Capture Group: greedy capture until the end of the line
            r'(.*)$'
        )
        current_type = re.search(input_extract_pattern, type_definition)
        if current_type is not None:
            self.log.debug(
                f'action=extract-type-from-definition, current-type={current_type.group(1)}'
            )
            return current_type.group(1).strip()
        return None

    def _extract_type_from_list(self, type_data: str) -> str:
        """Extract type data from Union[]."""
        # extract the type from List
        list_extract_pattern = r'^List\[(.*)\]'
        extract_from_list = re.search(list_extract_pattern, type_data)
        if extract_from_list is not None:
            self.log.debug(f'action=extract-type-from-list, type-data={type_data}')
            return extract_from_list.group(1)
        return type_data

    def _extract_type_from_method(self, type_data: str) -> str:
        """Extract type data from method (e.g. binary(min_length=2) -> binary)."""
        remove_args_pattern = r'\([^\)]*\)'
        type_data = re.sub(remove_args_pattern, '', type_data)
        self.log.debug(f'action=remove-args-pattern, types-data={type_data}')
        return type_data

    def _extract_type_from_optional(self, type_data: str) -> str:
        """Extract type data from Union[]."""
        optional_extract_pattern = r'Optional\[?(.*)\]'
        extracted_data = re.search(optional_extract_pattern, type_data)
        if extracted_data is not None:
            type_data = extracted_data.group(1)
            self.log.debug(f'action=extract-type-from-optional, type-data={type_data}')
            return extracted_data.group(1)
        return type_data

    def _extract_type_from_union(self, type_data: str) -> str:
        """Extract type data from Union[]."""
        union_extract_pattern = r'Union\[(\((.+)\)|.+)\]'
        extracted_data = re.search(union_extract_pattern, type_data)
        if extracted_data is not None:
            # return the last matched group that is not None
            type_data = [g for g in extracted_data.groups() if g is not None][-1]
            self.log.debug(f'action=extract-type-from-union, type-data={type_data}')
            return type_data
        return type_data

    def _generate_app_inputs_to_action(self):
        """Generate App Input dict from install.json and layout.json."""
        if self.ij.model.is_trigger_app is True:
            # Process the following App types:
            # - playbook trigger service Apps
            for ij_data in self.ij.model.params_dict.values():
                class_name = 'Trigger Config'
                if ij_data.service_config is True:
                    class_name = 'Service Config'
                self._add_input_to_action_class(False, ij_data, class_name)
        elif self.lj.has_layout is False or not self._tc_actions:
            # Process the following App types:
            # - Job Apps
            # - Playbook App with no layout.json
            # - Playbook App with layout.json but no tc_actions
            for ij_data in self.ij.model.params_dict.values():
                self._add_input_to_action_class(True, ij_data)
        else:
            # Process the following App types:
            # - Playbook App with layout.json and tc_action input
            for tc_action in self._tc_actions:
                if tc_action == 'Advanced Request':
                    # AdvancedRequestModel is included in tcex
                    continue

                for input_data in self.permutations.get_action_inputs(tc_action):
                    applies_to_all = self.permutations.get_input_applies_to_all(input_data.name)
                    self._add_input_to_action_class(applies_to_all, input_data, tc_action)

                    self.log.debug(
                        f'''action=inputs-to-action, input-name={input_data.name}, '''
                        f'''applies-to-all={applies_to_all}'''
                    )

    @staticmethod
    def _gen_tc_action_class_name(tc_action: str) -> str:
        """Format the action to a proper class name."""
        if tc_action is not None:
            # split to make pascal case
            _parts = [p.title() for p in tc_action.replace('_', ' ').split(' ')]

            # title case each word
            tc_action = ''.join([f'{p.title()}' for p in _parts]) + 'Model'

            # remove all non-alphanumeric characters and underscores
            tc_action = re.sub(r'[^a-zA-Z0-9]', '', tc_action)
        return tc_action

    def _gen_type(self, class_name: str, input_data: 'ParamsModel') -> str:
        """Determine the type value for the current input."""
        field_types = []

        # get map lookup key
        lookup_key = input_data.type
        required_key = 'optional' if input_data.required is False else 'required'
        if input_data.encrypt is True:
            lookup_key = 'Encrypt'

        # lookup input name in standards map to get pre-defined type.
        standard_name_type = self._standard_field_to_type_map(input_data.name)

        # get the type value and field type
        if standard_name_type is not None:
            type_ = standard_name_type.get('type')
            field_types.append(standard_name_type.get('field_type'))
        elif input_data.type == 'Boolean':
            type_ = 'bool'
        elif (
            input_data.encrypt is False
            and input_data.type == 'String'
            and input_data.playbook_data_type not in [[], None]
        ):
            type_, field_types = self._gen_type_from_playbook_data_type(
                required_key, input_data.playbook_data_type
            )
        else:
            try:
                type_ = self.input_static.type_map[lookup_key][required_key]['type']
                field_types = [self.input_static.type_map[lookup_key][required_key]['field_type']]
            except (AttributeError, KeyError) as ex:
                self.print_failure(f'Failed looking up type data for {input_data.type} ({ex}).')

        # wrap type data in Optional for non-required inputs
        if input_data.required is False and input_data.type not in ('Boolean'):
            type_ = f'Optional[{type_}]'

        # append default if one exists
        # if input_data.default is not None and input_data.type in ('Choice', 'String'):
        #     type_ += f' = \'{input_data.default}\''
        if input_data.type == 'Boolean':
            type_ += f' = {input_data.default or False}'

        # get the current value from the app_inputs.py file
        current_type = self._get_current_type(class_name, input_data.name)
        if current_type is not None and current_type != type_:
            self.report_mismatch.append(
                f'{input_data.name} -> calculated-type="{type_}" does '
                f'not match current-type="{current_type}"'
            )
            self.log.warning(
                f'input={input_data.name}, current-type={current_type}, '
                f'calculated-type={type_}, using=current-type'
            )
            field_types = self._get_current_field_types(current_type)
            type_ = current_type

        # add field types for import
        for field_type in field_types:
            if field_type is not None:
                # only add the field type if it is a defined field type
                if hasattr(FieldTypes, field_type):
                    self.field_type_modules.add(field_type)

        # add import data
        self._add_typing_import_module(type_)

        return type_

    def _gen_type_from_playbook_data_type(
        self, required_key: str, playbook_data_types: List[str]
    ) -> Tuple[str, List[str]]:
        """Return type based on playbook data type."""
        # TODO: what to do with Any Playbook Data Type?
        if 'Any' in playbook_data_types:
            self.typing_modules.add('Any')

        if len(playbook_data_types) == 1:
            _field_types = [
                self.input_static.type_map[playbook_data_types[0]][required_key]['field_type']
            ]
            _types = self.input_static.type_map[playbook_data_types[0]][required_key]['type']
        else:
            _types = []
            _field_types = []
            for lookup_key in playbook_data_types:
                _field_types.append(
                    self.input_static.type_map[lookup_key][required_key]['field_type']
                )
                _types.append(self.input_static.type_map[lookup_key][required_key]['type'])
            _types = f'''Union[{', '.join(_types)}]'''

        return _types, _field_types

    def _get_current_type(self, class_name: str, input_name: str) -> str:
        """Return the type from the current app_input.py file if found."""
        # Try to capture the value from the specific class first. If not
        # found, search the entire app_inputs.py file.
        type_definition = self.utils.find_line_in_code(
            needle=rf'\s+{input_name}: ',
            code=self.app_inputs_contents,
            trigger_start=rf'^class {class_name}',
            trigger_stop=r'^class ',
        ) or self.utils.find_line_in_code(needle=f'{input_name}: ', code=self.app_inputs_contents)
        # type_definition -> "string_encrypt: Optional[Sensitive]"
        self.log.debug(
            f'action=find-definition, input-name={input_name}, type-definition={type_definition}'
        )

        # parse out the actual type
        if type_definition is not None:
            current_type = self._extract_type_from_definition(input_name, type_definition)
            if current_type is not None:
                if 'Union' in current_type:
                    # the find method will return Union[(String, Optional[String])] that has
                    # the types as a tuple. this may be valid, but to keep the type
                    # consistent this bit of code will remove the extra ()/tuple. this
                    # needs to cover all cases, even the more complicated ones
                    # (e.g., "Optional[Union[(String, List[String])]]")
                    _types_data = self._extract_type_from_union(current_type)
                    union_original = f'Union[({_types_data})]'
                    union_replace = f'Union[{_types_data}]'
                    current_type = current_type.replace(union_original, union_replace)

                self.log.debug(
                    f'action=get-type, input-name={input_name}, current-type={current_type}'
                )
                return current_type

            self.log.warning(
                f'input found, but not matched: input={input_name}, '
                f'type_definition={type_definition}'
            )

        return None

    def _get_current_field_types(self, current_type_data: str) -> str:
        """Return the current type from the type data (e.g., integer(gt=2) -> integer)."""
        types_data = current_type_data

        # extract the type from Union (e.g. Union[binary(), List[binary()]])
        types_data = self._extract_type_from_union(types_data)

        # extract type from method (e.g. binary(min_length=2) -> binary)
        types_data = self._extract_type_from_method(types_data)

        types = []
        for type_ in types_data.split(', '):
            # extract the type from List[]
            type_ = self._extract_type_from_list(type_)

            # extract the type from Optional[]
            type_ = self._extract_type_from_optional(type_)

            # append types
            types.append(type_)

        self.log.debug(f'current-field-types={types}, current-type-data={current_type_data}')
        return types

    @staticmethod
    def _standard_field_to_type_map(input_name: str) -> str:
        """Return the type for "standard" input fields."""
        _field_name_to_type_map = {
            'confidence_rating': {
                'type': 'integer(ge=0, le=100)',
                'field_type': 'integer',
            },
            'last_run': {
                'type': 'DateTime',
                'field_type': 'DateTime',
            },
            'max_historical_poll_start': {
                'type': 'DateTime',
                'field_type': 'DateTime',
            },
            'poll_interval': {
                'type': 'integer(gt=0)',
                'field_type': 'integer',
            },
            'threat_rating': {
                'type': 'integer(ge=0, le=5)',
                'field_type': 'integer',
            },
        }
        return _field_name_to_type_map.get(input_name)

    @property
    def _tc_actions(self):
        """Return tc_action input valid values."""
        _tc_action = self.ij.model.get_param('tc_action')
        if not isinstance(_tc_action, NoneModel):
            return _tc_action.valid_values
        return []

    def _validator_always_array(self, always_array: List[str]) -> str:
        """Return code for always_array_validator."""
        _always_array = ', '.join(always_array)
        return [
            '',
            f'{self.i1}# ensure inputs that take single and array types always return an array',
            (
                f'{self.i1}_always_array = validator({_always_array}, '
                'allow_reuse=True)(always_array())'
            ),
        ]

    def _validator_always_array_check(self, input_data: 'ParamsModel') -> bool:
        """Return True if Single and Multiple types used."""
        array_type = False
        single_type = False
        for _type in input_data.playbook_data_type:
            if _type in self.utils.variable_playbook_array_types:
                array_type = True
            elif _type in self.utils.variable_playbook_single_types:
                single_type = True

        return all([array_type, single_type])

    def _validator_entity_input(self, entity_input: List[str]) -> str:
        """Return code for always_array_validator."""
        _entity_input = ', '.join(entity_input)
        return [
            '',
            f'{self.i1}# add entity_input validator for supported types',
            (
                f'''{self.i1}_entity_input = validator({_entity_input}, '''
                '''allow_reuse=True)(entity_input(only_field='value'))'''
            ),
        ]

    @staticmethod
    def _validator_entity_input_check(input_data: 'ParamsModel') -> bool:
        """Return True if Single and Multiple types used."""
        for _type in input_data.playbook_data_type:
            if _type in ['TCEntity', 'TCEntityArray']:
                return True
        return False

    @cached_property
    def app_inputs_contents(self):  # pylint: disable=no-self-use
        """Return app_inputs.py contents."""
        if os.path.isfile('app_inputs.py'):
            with open('app_inputs.py') as f:
                return f.read()
        return ''

    @property
    def app_inputs_data(self) -> dict:
        """Return base App inputs data."""
        if self._app_inputs_data is None:
            if self.ij.model.is_trigger_app:
                self._app_inputs_data = {'ServiceConfigModel': {}, 'TriggerConfigModel': {}}
            else:
                self._app_inputs_data = {'AppBaseModel': {}}

            # add a model for each action (for layout based Apps)
            self._add_action_classes()
        return self._app_inputs_data

    def generate(self):
        """Generate App Config File"""
        self.log.debug('--- generate: AppConfig ---')

        # generate the App Inputs
        self._generate_app_inputs_to_action()

        # generate input code first so that imports can be added
        _code_inputs = self._code_app_inputs_data

        # create the app_inputs.py code
        code = self.input_static.template_app_inputs_prefix(
            self.field_type_modules, self.pydantic_modules, self.typing_modules
        )
        code.extend(_code_inputs)
        if not isinstance(self.ij.model.get_param('tc_action'), NoneModel):
            # the App support tc_action and should use the tc_action input class
            self.typing_modules.add('Optional')
            code.extend(self.input_static.template_app_inputs_class_tc_action(self.class_model_map))
        else:
            code.extend(self.input_static.template_app_inputs_class)
        return code



"""Indicator / Indicators Model"""
# pylint: disable=no-member,no-self-argument,no-self-use,wrong-import-position
# standard library
from datetime import datetime
from typing import List, Optional

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.utils import Utils


class IndicatorsModel(
    BaseModel,
    title='Indicators Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Indicators Model"""

    _mode_support = PrivateAttr(True)

    data: Optional[List['IndicatorModel']] = Field(
        [],
        description='The data for the Indicators.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


class IndicatorDataModel(
    BaseModel,
    title='Indicator Data Model',
    alias_generator=Utils().snake_to_camel,
    validate_assignment=True,
):
    """Indicators Data Model"""

    data: Optional[List['IndicatorModel']] = Field(
        [],
        description='The data for the Indicators.',
        methods=['POST', 'PUT'],
        title='data',
    )


class IndicatorModel(
    V3ModelABC,
    alias_generator=Utils().snake_to_camel,
    extra=Extra.allow,
    title='Indicator Model',
    validate_assignment=True,
):
    """Indicator Model"""

    _associated_type = PrivateAttr(True)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    active: bool = Field(
        None,
        description='Is the indicator active?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='active',
    )
    active_locked: bool = Field(
        None,
        description='Lock the indicator active value?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='activeLocked',
    )
    address: Optional[str] = Field(
        None,
        applies_to=['EmailAddress'],
        conditional_required=['EmailAddress'],
        description=(
            'The email address associated with this indicator (EmailAddress specific summary '
            'field).'
        ),
        methods=['POST'],
        read_only=False,
        title='address',
    )
    associated_artifacts: Optional['ArtifactsModel'] = Field(
        None,
        description='A list of Artifacts associated with this Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedArtifacts',
    )
    associated_cases: Optional['CasesModel'] = Field(
        None,
        description='A list of Cases associated with this Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedCases',
    )
    associated_groups: Optional['GroupsModel'] = Field(
        None,
        description='A list of groups that this indicator is associated with.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: Optional['IndicatorsModel'] = Field(
        None,
        description='A list of indicators associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    attributes: Optional['IndicatorAttributesModel'] = Field(
        None,
        description='A list of Attributes corresponding to the Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='attributes',
    )
    confidence: Optional[int] = Field(
        None,
        description='The indicator threat confidence.',
        methods=['POST', 'PUT'],
        maximum=100,
        minimum=0,
        read_only=False,
        title='confidence',
    )
    date_added: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    description: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The indicator description text.',
        read_only=True,
        title='description',
    )
    dns_active: bool = Field(
        None,
        applies_to=['Host'],
        description='Is dns active for the indicator?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='dnsActive',
    )
    dns_resolution: Optional[dict] = Field(
        None,
        allow_mutation=False,
        applies_to=['Host', 'Address'],
        conditional_required=['Host', 'Address'],
        description='Dns resolution data for the Host or Address indicator.',
        read_only=True,
        title='dnsResolution',
    )
    enrichment: Optional[dict] = Field(
        None,
        allow_mutation=False,
        description='Enrichment data.',
        read_only=True,
        title='enrichment',
    )
    false_positive_reported_by_user: bool = Field(
        None,
        allow_mutation=False,
        description='Has a false positive been reported by this user for this indicator today?',
        read_only=True,
        title='falsePositiveReportedByUser',
    )
    false_positives: Optional[int] = Field(
        None,
        allow_mutation=False,
        description='The number of false positives reported for this indicator.',
        read_only=True,
        title='falsePositives',
    )
    file_actions: Optional['FileActionsModel'] = Field(
        None,
        description='The type of file action associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileActions',
    )
    file_occurrences: Optional['FileOccurrencesModel'] = Field(
        None,
        description='A list of file occurrences associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileOccurrences',
    )
    geo_location: Optional[dict] = Field(
        None,
        allow_mutation=False,
        applies_to=['Host', 'Address'],
        conditional_required=['Host', 'Address'],
        description='Geographical localization of the Host or Address indicator.',
        read_only=True,
        title='geoLocation',
    )
    host_name: Optional[str] = Field(
        None,
        applies_to=['Host'],
        conditional_required=['Host'],
        description='The host name of the indicator (Host specific summary field).',
        methods=['POST'],
        read_only=False,
        title='hostName',
    )
    id: Optional[int] = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    investigation_links: Optional[dict] = Field(
        None,
        allow_mutation=False,
        description=(
            'Resource links that provide additional information to assist in investigation.'
        ),
        read_only=True,
        title='investigationLinks',
    )
    ip: Optional[str] = Field(
        None,
        applies_to=['Address'],
        conditional_required=['Address'],
        description=(
            'The ip address associated with this indicator (Address specific summary field).'
        ),
        methods=['POST'],
        read_only=False,
        title='ip',
    )
    last_false_positive: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The date and time of the last false positive reported for this indicator.',
        read_only=True,
        title='lastFalsePositive',
    )
    last_modified: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The date and time that the indicator was last modified.',
        read_only=True,
        title='lastModified',
    )
    last_observed: Optional[datetime] = Field(
        None,
        allow_mutation=False,
        description='The date and time that the indicator was last observed.',
        read_only=True,
        title='lastObserved',
    )
    legacy_link: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='A link to the legacy ThreatConnect details page for this entity.',
        read_only=True,
        title='legacyLink',
    )
    md5: Optional[str] = Field(
        None,
        applies_to=['File'],
        description='The md5 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='md5',
    )
    mode: Optional[str] = Field(
        None,
        applies_to=['File'],
        description='The operation to perform on the file hashes (delete | merge).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='mode',
    )
    observations: Optional[int] = Field(
        None,
        allow_mutation=False,
        description='The number of times this indicator has been observed.',
        read_only=True,
        title='observations',
    )
    owner_id: Optional[int] = Field(
        None,
        description='The id of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerId',
    )
    owner_name: Optional[str] = Field(
        None,
        description='The name of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerName',
    )
    private_flag: bool = Field(
        None,
        description='Is this indicator private?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='privateFlag',
    )
    rating: Optional[int] = Field(
        None,
        description='The indicator threat rating.',
        methods=['POST', 'PUT'],
        maximum=5,
        minimum=0,
        read_only=False,
        title='rating',
    )
    security_labels: Optional['SecurityLabelsModel'] = Field(
        None,
        description=(
            'A list of Security Labels corresponding to the Intel item (NOTE: Setting this '
            'parameter will replace any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityLabels',
    )
    sha1: Optional[str] = Field(
        None,
        applies_to=['File'],
        description='The sha1 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='sha1',
    )
    sha256: Optional[str] = Field(
        None,
        applies_to=['File'],
        description='The sha256 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='sha256',
    )
    size: Optional[int] = Field(
        None,
        applies_to=['File'],
        description='The size of the file.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='size',
    )
    source: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The source for this indicator.',
        read_only=True,
        title='source',
    )
    summary: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='The indicator summary.',
        read_only=True,
        title='summary',
    )
    tags: Optional['TagsModel'] = Field(
        None,
        description=(
            'A list of Tags corresponding to the item (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    text: Optional[str] = Field(
        None,
        applies_to=['URL'],
        conditional_required=['URL'],
        description='The url text value of the indicator (Url specific summary field).',
        methods=['POST'],
        read_only=False,
        title='text',
    )
    threat_assess_confidence: Optional[float] = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess confidence for this indicator.',
        read_only=True,
        title='threatAssessConfidence',
    )
    threat_assess_rating: Optional[float] = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess rating for this indicator.',
        read_only=True,
        title='threatAssessRating',
    )
    threat_assess_score: Optional[int] = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score for this indicator.',
        read_only=True,
        title='threatAssessScore',
    )
    threat_assess_score_false_positive: Optional[int] = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score for false positives related to this indicator.',
        read_only=True,
        title='threatAssessScoreFalsePositive',
    )
    threat_assess_score_observed: Optional[int] = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score observed for this indicator.',
        read_only=True,
        title='threatAssessScoreObserved',
    )
    type: Optional[str] = Field(
        None,
        description='The **type** for the Indicator.',
        methods=['POST', 'PUT'],
        min_length=1,
        read_only=False,
        title='type',
    )
    value1: Optional[str] = Field(
        None,
        description='Custom Indicator summary field value1.',
        methods=['POST'],
        read_only=False,
        title='value1',
    )
    value2: Optional[str] = Field(
        None,
        description='Custom Indicator summary field value2.',
        methods=['POST'],
        read_only=False,
        title='value2',
    )
    value3: Optional[str] = Field(
        None,
        description='Custom Indicator summary field value3.',
        methods=['POST'],
        read_only=False,
        title='value3',
    )
    web_link: Optional[str] = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    whois: Optional[dict] = Field(
        None,
        allow_mutation=False,
        applies_to=['Host'],
        conditional_required=['Host'],
        description='The whois data for the indicator.',
        read_only=True,
        title='whois',
    )
    whois_active: bool = Field(
        None,
        applies_to=['Host'],
        description='Is whois active for the indicator?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='whoisActive',
    )

    @validator('associated_artifacts', always=True)
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()
        return v

    @validator('associated_cases', always=True)
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()
        return v

    @validator('file_actions', always=True)
    def _validate_file_actions(cls, v):
        if not v:
            return FileActionsModel()
        return v

    @validator('file_occurrences', always=True)
    def _validate_file_occurrences(cls, v):
        if not v:
            return FileOccurrencesModel()
        return v

    @validator('associated_groups', always=True)
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()
        return v

    @validator('attributes', always=True)
    def _validate_indicator_attributes(cls, v):
        if not v:
            return IndicatorAttributesModel()
        return v

    @validator('associated_indicators', always=True)
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()
        return v

    @validator('security_labels', always=True)
    def _validate_security_labels(cls, v):
        if not v:
            return SecurityLabelsModel()
        return v

    @validator('tags', always=True)
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()
        return v


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.file_actions.file_action_model import FileActionsModel
from tcex.api.tc.v3.file_occurrences.file_occurrence_model import FileOccurrencesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicator_attributes.indicator_attribute_model import IndicatorAttributesModel
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelsModel
from tcex.api.tc.v3.tags.tag_model import TagsModel

# add forward references
IndicatorDataModel.update_forward_refs()
IndicatorModel.update_forward_refs()
IndicatorsModel.update_forward_refs()



"""TcEx Generate Configurations CLI Command"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.app_config.models import InstallJsonModel
from tcex.bin.bin_abc import BinABC

if TYPE_CHECKING:
    # first-party
    from tcex.app_config import AppSpecYml


class SpecToolInstallJson(BinABC):
    """Generate App Config File"""

    def __init__(self, asy: 'AppSpecYml'):
        """Initialize class properties."""
        super().__init__()
        self.asy = asy

        # properties
        self.filename = 'install.json'

    def _add_standard_fields(self, install_json_data: dict):
        """Add field that apply to ALL App types."""
        install_json_data.update(
            {
                'allowOnDemand': self.asy.model.allow_on_demand,
                'apiUserTokenParam': self.asy.model.api_user_token_param,
                'appId': self.asy.model.app_id,
                'category': self.asy.model.category,
                'deprecatesApps': self.asy.model.deprecates_apps,
                'displayName': self.asy.model.display_name,
                'features': self.asy.model.features,
                'labels': self.asy.model.labels,
                'languageVersion': self.asy.model.language_version,
                'listDelimiter': self.asy.model.list_delimiter,
                'minServerVersion': str(self.asy.model.min_server_version),
                'params': [p.dict(by_alias=True) for p in self.asy.model.params],
                'programLanguage': self.asy.model.program_language,
                'programMain': self.asy.model.program_main,
                # 'programName': '',
                'programVersion': str(self.asy.model.program_version),
                'runtimeLevel': self.asy.model.runtime_level,
            }
        )

    def _add_note(self, install_json_data: dict):
        """Add top level note to install.json."""
        _note = f'{self.asy.model.note}{self._note_per_action}\n'

        # some Apps have a serviceDetails section which needs to be appended to the App notes
        if self.asy.model.service_details is not None:
            _note += self.asy.model.service_details

        install_json_data['note'] = _note

    def _add_type_api_service_fields(self, install_json_data: dict):
        """Add field that apply to ALL App types."""
        if self.asy.model.is_api_service_app:
            install_json_data['displayPath'] = self.asy.model.display_path

    def _add_type_organization_fields(self, install_json_data: dict):
        """Add field that apply to ALL App types."""
        if self.asy.model.is_feed_app:
            # the nested job object is not part of the install.json, it
            # instead gets written to the *.job.json file.
            if self.asy.model.organization.feeds:
                _feeds = []
                for feed in self.asy.model.organization.feeds:
                    feed = feed.dict(by_alias=True)
                    if feed.get('job') is not None:
                        del feed['job']
                    _feeds.append(feed)
                install_json_data['feeds'] = _feeds

            # publish_out_files
            _publish_out_files = self.asy.model.organization.publish_out_files
            if _publish_out_files:
                install_json_data['publishOutFiles'] = _publish_out_files

            # repeating_minutes
            _repeating_minutes = self.asy.model.organization.repeating_minutes
            if _repeating_minutes:
                install_json_data['repeatingMinutes'] = _repeating_minutes

    def _add_type_playbook_fields(self, install_json_data: dict):
        """Add field that apply to ALL App types."""
        if self.asy.model.is_playbook_app or self.asy.model.is_trigger_app:
            install_json_data['allowRunAsUser'] = self.asy.model.allow_run_as_user
            install_json_data['playbook'] = {
                'outputPrefix': self.asy.model.output_prefix,
                'outputVariables': [
                    ov.dict(by_alias=True) for ov in self.asy.model.output_variables
                ],
                'type': self.asy.model.category,
            }
            if self.asy.model.playbook and self.asy.model.playbook.retry.allowed is True:
                install_json_data['playbook']['retry'] = self.asy.model.playbook.retry.dict(
                    by_alias=True
                )

    @property
    def _note_per_action(self):
        """Return note per action string."""
        # note per action is only supported on playbook Apps
        note_per_action = ''
        if self.asy.model.note_per_action:
            note_per_action = '\n\n'.join(self.asy.model.note_per_action_formatted)
        return note_per_action

    def generate(self):
        """Generate the install.json file data."""
        # all keys added to dict must be in camelCase
        install_json_data = {}

        # add standard fields
        self._add_standard_fields(install_json_data)

        # add note
        self._add_note(install_json_data)

        # add app type api service fields
        self._add_type_api_service_fields(install_json_data)

        # add app type organization fields
        self._add_type_organization_fields(install_json_data)

        # add app type organization fields
        self._add_type_playbook_fields(install_json_data)

        # update sequence numbers
        for sequence, param in enumerate(install_json_data.get('params'), start=1):
            param['sequence'] = sequence

        return InstallJsonModel(**install_json_data)

    @staticmethod
    def generate_schema():
        """Return the schema for the install.json file."""
        return InstallJsonModel.schema_json(indent=2, sort_keys=True)



"""TcEx Framework module init file."""
# flake8: noqa
# standard library
import logging

# first-party
from tcex.logger.cache_handler import CacheHandler  # pylint: disable=no-name-in-module
from tcex.logger.trace_logger import TraceLogger  # pylint: disable=no-name-in-module

# init logger before instantiating tcex
logging.setLoggerClass(TraceLogger)
logger = logging.getLogger('tcex')
logger.setLevel(logging.TRACE)  # pylint: disable=E1101

# add TEMP cache handler, which will be removed in tcex.py (we don't know log path here)
cache = CacheHandler()
cache.set_name('cache')
cache.setLevel(logging.TRACE)  # pylint: disable=no-member
cache.setFormatter(
    '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
    '(%(filename)s:%(funcName)s:%(lineno)d)'
)
logger.addHandler(cache)

# pylint: disable=wrong-import-position
from .__metadata__ import (
    __author__,
    __author_email__,
    __description__,
    __license__,
    __package_name__,
    __url__,
    __version__,
)

try:
    # first-party
    from tcex.tcex import TcEx
except ImportError as e:
    print(f'Error: {e}')
    print('Try running tclib')
    raise



"""TcEx JSON Model"""
# pylint: disable=no-self-argument,no-self-use; noqa: N805
# standard library
from enum import Enum
from pathlib import PosixPath
from typing import List, Optional

# third-party
from pydantic import BaseModel, validator

# first-party
from tcex.pleb.env_path import EnvPath

__all__ = ['TcexJsonModel']


class LibVersionModel(BaseModel):
    """Model for tcex_json.lib_version"""

    lib_dir: EnvPath
    python_executable: EnvPath

    class Config:
        """DataModel Config"""

        validate_assignment = True


class PackageModel(BaseModel):
    """Model for tcex_json.package"""

    app_name: str
    app_version: Optional[str]
    excludes: list
    output_dir: str = 'target'

    @validator('excludes')
    def sorted(cls, v):
        """Change value for excludes field."""
        # the requirements.txt file is required for App Builder
        v = [e for e in v if e != 'requirements.txt']
        return list(sorted(set(v)))

    class Config:
        """DataModel Config"""

        validate_assignment = True


class TemplateTypes(str, Enum):
    """Enum for tcex.template_type"""

    api_service = 'api_service'
    external = 'external'
    organization = 'organization'
    playbook = 'playbook'
    trigger_service = 'trigger_service'
    web_api_service = 'web_api_service'
    webhook_trigger_service = 'webhook_trigger_service'


class TcexJsonModel(BaseModel):
    """TcEx JSON Model"""

    lib_versions: Optional[List[LibVersionModel]]
    package: PackageModel
    template_name: Optional[str] = None
    template_repo_hash: Optional[str] = None
    template_type: Optional[TemplateTypes] = None

    class Config:
        """DataModel Config"""

        json_encoders = {PosixPath: lambda v: v.original_value}
        use_enum_values = True
        validate_assignment = True



"""TcEx Generate Configurations CLI Command"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.app_config import TcexJson
from tcex.app_config.models.job_json_model import JobJsonModel
from tcex.bin.bin_abc import BinABC

if TYPE_CHECKING:
    # first-party
    from tcex.app_config import AppSpecYml


class SpecToolJobJson(BinABC):
    """Generate App Config File"""

    def __init__(self, asy: 'AppSpecYml'):
        """Initialize class properties."""
        super().__init__()
        self.asy = asy

        # properties
        self.tj = TcexJson(logger=self.log)

    def generate(self):
        """Generate the layout.json file data."""
        if self.asy.model.is_feed_app and self.asy.model.organization.feeds:
            for feed in self.asy.model.organization.feeds:
                _job_data = feed.job.dict(by_alias=True)
                app_name = self.tj.model.package.app_name.replace('_', ' ')

                # handle statically defined version in tcex.json file
                version = f'v{self.asy.model.program_version.major}'
                if self.tj.model.package.app_version:
                    version = self.tj.model.package.app_version

                _job_data['programName'] = f'{app_name} {version}'
                _job_data['programVersion'] = str(self.asy.model.program_version)
                yield feed.job_file, JobJsonModel(**_job_data)

    @staticmethod
    def generate_schema():
        """Return the schema for the install.json file."""
        return JobJsonModel.schema_json(indent=2, sort_keys=True)



"""TcEx JSON App Config"""
# standard library
import json
import logging
import os
from collections import OrderedDict
from pathlib import Path
from typing import Optional

# first-party
from tcex.app_config.models import JobJsonModel
from tcex.backports import cached_property
from tcex.pleb.singleton import Singleton

# get tcex logger
tcex_logger = logging.getLogger('tcex')


class JobJson(metaclass=Singleton):
    """Provide a model for the tcex.json config file."""

    def __init__(
        self,
        filename: Optional[str] = None,
        path: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize class properties."""
        filename = filename or 'tcex.json'
        path = path or os.getcwd()
        self.log = logger or tcex_logger

        # properties
        self.fqfn = Path(os.path.join(path, filename))

    @cached_property
    def contents(self) -> dict:
        """Return tcex.json file contents."""
        _contents = {}

        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    _contents = json.load(fh, object_pairs_hook=OrderedDict)
            except OSError:  # pragma: no cover
                self.log.error(
                    f'feature=tcex-json, exception=failed-reading-file, filename={self.fqfn}'
                )
        else:  # pragma: no cover
            self.log.error(f'feature=tcex-json, exception=file-not-found, filename={self.fqfn}')

        return _contents

    @cached_property
    def model(self) -> 'JobJsonModel':
        """Return the Install JSON model."""
        return JobJsonModel(**self.contents)

    # TODO: [low] possibly add auto fix of version and program name and then uncomment this code.
    # def write(self):
    #     """Write current data file."""
    #     data = self.model.json(
    #         by_alias=True, exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True
    #     )
    #     with self.fqfn.open(mode='w') as fh:
    #         fh.write(f'{data}\n')



"""Bin module for TcEx Framework"""
# flake8: noqa
# first-party
from tcex.bin.dep import Dep
from tcex.bin.deploy import Deploy
from tcex.bin.package import Package
from tcex.bin.spec_tool import SpecTool
from tcex.bin.template import Template
from tcex.bin.validate import Validate



"""API"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc import TC
from tcex.backports import cached_property

if TYPE_CHECKING:
    # third-party
    from requests import Session

    # first-party
    from tcex.input.input import Input


class API:
    """API

    Args:
        inputs: An instance of the Input class.
        session_tc: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, inputs: 'Input', session_tc: 'Session'):
        """Initialize Class properties."""
        self.inputs = inputs
        self.session_tc = session_tc

    @cached_property
    def tc(self) -> 'TC':
        """Return a case management instance."""
        return TC(self.inputs, self.session_tc)



"""Batch"""
# flake8: noqa
# first-party
from tcex.api.tc.v2.batch.batch import Batch
from tcex.api.tc.v2.batch.batch_submit import BatchSubmit
from tcex.api.tc.v2.batch.batch_writer import BatchWriter, GroupType, IndicatorType



"""TcEx Framework"""

# standard library
import inspect
import logging
import os
import platform
import signal
import threading
from typing import TYPE_CHECKING, Dict, Optional, Union

# third-party
from requests import Session

# first-party
from tcex.api import API
from tcex.api.tc.utils.threat_intel_utils import ThreatIntelUtils
from tcex.api.tc.v2.v2 import V2
from tcex.api.tc.v3.v3 import V3
from tcex.app_config.install_json import InstallJson
from tcex.app_feature import AdvancedRequest
from tcex.backports import cached_property
from tcex.exit.exit import ExitCode, ExitService
from tcex.input.input import Input
from tcex.key_value_store import KeyValueApi, KeyValueMock, KeyValueRedis, RedisClient
from tcex.logger.logger import Logger  # pylint: disable=no-name-in-module
from tcex.playbook import Playbook
from tcex.pleb.proxies import proxies
from tcex.pleb.registry import registry
from tcex.pleb.scoped_property import scoped_property
from tcex.sessions.auth.tc_auth import TcAuth
from tcex.sessions.external_session import ExternalSession
from tcex.sessions.tc_session import TcSession
from tcex.tokens import Tokens
from tcex.utils import Utils
from tcex.utils.file_operations import FileOperations

if TYPE_CHECKING:
    # first-party
    from tcex.logger.trace_logger import TraceLogger  # pylint: disable=no-name-in-module
    from tcex.services.api_service import ApiService
    from tcex.services.common_service_trigger import CommonServiceTrigger
    from tcex.services.webhook_trigger_service import WebhookTriggerService
    from tcex.sessions.auth.hmac_auth import HmacAuth
    from tcex.sessions.auth.token_auth import TokenAuth


class TcEx:
    """Provides basic functionality for all types of TxEx Apps.

    Args:
        config (dict, kwargs): A dictionary containing configuration items typically used by
            external Apps.
        config_file (str, kwargs): A filename containing JSON configuration items typically used
            by external Apps.
    """

    def __init__(self, **kwargs):
        """Initialize Class Properties."""
        # catch interrupt signals specifically based on thread name
        signal.signal(signal.SIGINT, self._signal_handler)
        if platform.system() != 'Windows':
            signal.signal(signal.SIGHUP, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

        # Property defaults
        self._config: dict = kwargs.get('config') or {}
        self._log = None
        self._jobs = None
        self._redis_client = None
        self._service = None
        self.ij = InstallJson()
        self.main_os_pid = os.getpid()

        # init inputs
        self.inputs = Input(self._config, kwargs.get('config_file'))

        # add methods to registry
        registry.add_method(self.inputs.resolve_variable)

        # add methods to registry
        registry.register(self)
        registry.add_service(Input, self.inputs)

        # log standard App info early so it shows at the top of the logfile
        self.logger.log_info(self.inputs.model_unresolved)

    def _signal_handler(self, signal_interrupt: int, _):
        """Handle signal interrupt."""
        call_file: str = os.path.basename(inspect.stack()[1][0].f_code.co_filename)
        call_module: str = inspect.stack()[1][0].f_globals['__name__'].lstrip('Functions.')
        call_line: int = inspect.stack()[1][0].f_lineno
        self.log.error(
            f'App interrupted - file: {call_file}, method: {call_module}, line: {call_line}.'
        )
        exit_code = ExitCode.SUCCESS
        if threading.current_thread().name == 'MainThread' and signal_interrupt in (2, 15):
            exit_code = ExitCode.FAILURE

        self.exit(exit_code, 'The App received an interrupt signal and will now exit.')

    @property
    def _user_agent(self):
        """Return a User-Agent string."""
        return {
            'User-Agent': (
                f'TcEx/{__import__(__name__).__version__}, '
                f'{self.ij.model.display_name}/{self.ij.model.program_version}'
            )
        }

    def advanced_request(
        self,
        session: Session,
        output_prefix: str,
        timeout: Optional[int] = 600,
    ) -> 'AdvancedRequest':
        """Return instance of AdvancedRequest.

        Args:
            session: An instance of requests.Session.
            output_prefix: A value to prepend to outputs.
            timeout: The number of second before timing out the request.
        """
        return AdvancedRequest(self.inputs, self.playbook, session, output_prefix, timeout)

    @property
    def api(self) -> 'API':
        """Return instance of Threat Intel Utils."""
        return API(self.inputs, self.session_tc)

    def exit(self, code: Optional[ExitCode] = None, msg: Optional[str] = None):
        """Application exit method with proper exit code

        The method will run the Python standard sys.exit() with the exit code
        previously defined via :py:meth:`~tcex.tcex.TcEx.exit_code` or provided
        during the call of this method.

        Args:
            code: The exit code value for the app.
            msg: A message to log and add to message tc output.
        """
        # get correct code
        self.exit_service.exit(code, msg)  # pylint: disable=no-member

    @property
    def exit_code(self) -> 'ExitCode':
        """Return the current exit code."""
        return self.exit_service.exit_code  # pylint: disable=no-member

    @exit_code.setter
    def exit_code(self, code: 'ExitCode'):
        """Set the App exit code.

        For TC Exchange Apps there are 3 supported exit codes.
        * 0 indicates a normal exit
        * 1 indicates a failure during execution
        * 3 indicates a partial failure

        Args:
            code (int): The exit code value for the app.
        """
        self.exit_service.exit_code = code

    @registry.factory(ExitService)
    @scoped_property
    def exit_service(self) -> 'ExitService':
        """Return an ExitService object."""
        # TODO: [high] @cblades - inputs being required for exit prevents AOT from exiting
        return self.get_exit_service(self.inputs)

    @cached_property
    def file_operations(self) -> 'FileOperations':  # pylint: disable=no-self-use
        """Include the Utils module."""
        return FileOperations(
            out_path=self.inputs.model_unresolved.tc_out_path,
            temp_path=self.inputs.model_unresolved.tc_temp_path,
        )

    @staticmethod
    def get_exit_service(inputs) -> 'ExitService':
        """Create an ExitService object."""
        return ExitService(inputs)

    def get_playbook(
        self, context: Optional[str] = None, output_variables: Optional[list] = None
    ) -> 'Playbook':
        """Return a new instance of playbook module.

        Args:
            context: The KV Store context/session_id. For PB Apps the context is provided on
                startup, but for service Apps each request gets a different context.
            output_variables: The requested output variables. For PB Apps outputs are provided on
                startup, but for service Apps each request gets different outputs.
        """
        return Playbook(self.key_value_store, context, output_variables)

    @staticmethod
    def get_redis_client(
        host: str, port: int, db: int = 0, blocking_pool: bool = False, **kwargs
    ) -> 'RedisClient':
        """Return a *new* instance of Redis client.

        For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

        Args:
            host: The REDIS host. Defaults to localhost.
            port: The REDIS port. Defaults to 6379.
            db: The REDIS db. Defaults to 0.
            blocking_pool: Use BlockingConnectionPool instead of ConnectionPool.
            errors (str, kwargs): The REDIS errors policy (e.g. strict).
            max_connections (int, kwargs): The maximum number of connections to REDIS.
            password (str, kwargs): The REDIS password.
            socket_timeout (int, kwargs): The REDIS socket timeout.
            timeout (int, kwargs): The REDIS Blocking Connection Pool timeout value.

        Returns:
            Redis.client: An instance of redis client.
        """
        return RedisClient(
            host=host, port=port, db=db, blocking_pool=blocking_pool, **kwargs
        ).client

    def get_session_tc(
        self,
        auth: Optional[Union['HmacAuth', 'TokenAuth', 'TcAuth']] = None,
        base_url: Optional[str] = None,
        log_curl: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,  # pylint: disable=redefined-outer-name
        proxies_enabled: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> 'TcSession':
        """Return an instance of Requests Session configured for the ThreatConnect API.

        No args are required to get a working instance of TC Session instance.

        This method allows for getting a new instance of TC Session instance. This can be
        very useful when connecting between multiple TC instances (e.g., migrating data).
        """
        if log_curl is None:
            log_curl = self.inputs.model_unresolved.tc_log_curl

        if proxies_enabled is None:
            proxies_enabled = self.inputs.model_unresolved.tc_proxy_tc

        if verify is None:
            verify = self.inputs.model_unresolved.tc_verify

        if self.ij.is_external_app is True:
            auth = auth or TcAuth(
                tc_api_access_id=self.inputs.model_unresolved.tc_api_access_id,
                tc_api_secret_key=self.inputs.model_unresolved.tc_api_secret_key,
            )
        else:
            auth = auth or TcAuth(
                tc_api_access_id=self.inputs.model_unresolved.tc_api_access_id,
                tc_api_secret_key=self.inputs.model_unresolved.tc_api_secret_key,
                tc_token=self.token,
            )

        return TcSession(
            auth=auth,
            base_url=base_url or self.inputs.model_unresolved.tc_api_path,
            log_curl=log_curl,
            proxies=proxies or self.proxies,
            proxies_enabled=proxies_enabled,
            user_agent=self._user_agent,
            verify=verify,
        )

    def get_session_external(self) -> 'ExternalSession':
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        _session_external = ExternalSession()

        # add User-Agent to headers
        _session_external.headers.update(self._user_agent)

        # add proxy support if requested
        if self.inputs.model_unresolved.tc_proxy_external:
            _session_external.proxies = self.proxies
            self.log.info(
                f'Using proxy host {self.inputs.model_unresolved.tc_proxy_host}:'
                f'{self.inputs.model_unresolved.tc_proxy_port} for external session.'
            )

        if self.inputs.model_unresolved.tc_log_curl:
            _session_external.log_curl = True

        return _session_external

    # def get_ti(self) -> 'ThreatIntelligence':
    #     """Include the Threat Intel Module."""
    #     return ThreatIntelligence(session=self.get_session_tc())

    @registry.factory('KeyValueStore')
    @scoped_property
    def key_value_store(self) -> Union['KeyValueApi', 'KeyValueRedis']:
        """Return the correct KV store for this execution.

        The TCKeyValueAPI KV store is limited to two operations (create and read),
        while the Redis kvstore wraps a few other Redis methods.
        """
        if self.inputs.model_unresolved.tc_kvstore_type == 'Redis':
            return KeyValueRedis(self.redis_client)

        if self.inputs.model_unresolved.tc_kvstore_type == 'TCKeyValueAPI':
            return KeyValueApi(self.session_tc)

        if self.inputs.model_unresolved.tc_kvstore_type == 'Mock':
            self.log.warning(
                'Using mock key-value store.  '
                'This should *never* happen when running in-platform.'
            )
            return KeyValueMock()

        raise RuntimeError(
            f'Invalid KV Store Type: ({self.inputs.model_unresolved.tc_kvstore_type})'
        )

    @property
    def log(self) -> 'TraceLogger':
        """Return a valid logger."""
        if self._log is None:
            self._log = self.logger.log
        return self._log

    @log.setter
    def log(self, log: object):
        """Return a valid logger."""
        if isinstance(log, logging.Logger):
            self._log = log

    @cached_property
    def logger(self) -> 'Logger':
        """Return logger."""
        _logger = Logger(logger_name='tcex')

        # set logger to prevent recursion issue on get_session_tc
        self._log = _logger.log

        # add api handler
        if (
            self.inputs.contents.get('tc_token') is not None
            and self.inputs.contents.get('tc_log_to_api') is True
        ):
            _logger.add_api_handler(
                session_tc=self.get_session_tc(), level=self.inputs.model_unresolved.tc_log_level
            )

        # add rotating log handler
        _logger.add_rotating_file_handler(
            name='rfh',
            filename=self.inputs.model_unresolved.tc_log_file,
            path=self.inputs.model_unresolved.tc_log_path,
            backup_count=self.inputs.model_unresolved.tc_log_backup_count,
            max_bytes=self.inputs.model_unresolved.tc_log_max_bytes,
            level=self.inputs.model_unresolved.tc_log_level,
        )

        # set logging level
        _logger.update_handler_level(level=self.inputs.model_unresolved.tc_log_level)
        _logger.log.setLevel(_logger.log_level(self.inputs.model_unresolved.tc_log_level))

        # replay cached log events
        _logger.replay_cached_events(handler_name='cache')

        return _logger

    @registry.factory(Playbook)
    @scoped_property
    def playbook(self) -> 'Playbook':
        """Return an instance of Playbooks module.

        This property defaults context and outputvariables to arg values.

        Returns:
            tcex.playbook.Playbooks: An instance of Playbooks
        """
        return self.get_playbook(
            context=self.inputs.model_unresolved.tc_playbook_kvstore_context,
            output_variables=self.inputs.model_unresolved.tc_playbook_out_variables,
        )

    @cached_property
    def proxies(self) -> dict:
        """Format the proxy configuration for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {"http": "http://user:pass@10.10.1.10:3128/"}

        Returns:
           (dict): Dictionary of proxy settings
        """
        return proxies(
            proxy_host=self.inputs.model_unresolved.tc_proxy_host,
            proxy_port=self.inputs.model_unresolved.tc_proxy_port,
            proxy_user=self.inputs.model_unresolved.tc_proxy_username,
            proxy_pass=self.inputs.model_unresolved.tc_proxy_password,
        )

    @registry.factory(RedisClient)
    @scoped_property
    def redis_client(self) -> 'RedisClient':
        """Return redis client instance configure for Playbook/Service Apps."""
        return self.get_redis_client(
            host=self.inputs.contents.get('tc_kvstore_host'),
            port=self.inputs.contents.get('tc_kvstore_port'),
            db=0,
        )

    def results_tc(self, key: str, value: str):
        """Write data to results_tc file in TcEX specified directory.

        The TcEx platform support persistent values between executions of the App.  This
        method will store the values for TC to read and put into the Database.

        Args:
            key: The data key to be stored.
            value: The data value to be stored.
        """
        if os.access(self.inputs.model_unresolved.tc_out_path, os.W_OK):
            results_file = f'{self.inputs.model_unresolved.tc_out_path}/results.tc'
        else:
            results_file = 'results.tc'

        new = True
        # ensure file exists
        open(results_file, 'a').close()  # pylint: disable=consider-using-with
        with open(results_file, 'r+') as fh:
            results = ''
            for line in fh.read().strip().split('\n'):
                if not line:
                    continue
                try:
                    k, v = line.split(' = ')
                except ValueError:
                    # handle null/empty value (e.g., "name =")
                    k, v = line.split(' =')
                if k == key:
                    v = value
                    new = False
                if v is not None:
                    results += f'{k} = {v}\n'
            if new and value is not None:  # indicates the key/value pair didn't already exist
                results += f'{key} = {value}\n'
            fh.seek(0)
            fh.write(results)
            fh.truncate()

    @cached_property
    def service(self) -> Union['ApiService', 'CommonServiceTrigger', 'WebhookTriggerService']:
        """Include the Service Module."""
        if self.ij.model.is_api_service_app:
            from .services import ApiService as Service
        elif self.ij.model.is_trigger_app and not self.ij.model.is_webhook_trigger_app:
            from .services import CommonServiceTrigger as Service
        elif self.ij.model.is_webhook_trigger_app:
            from .services import WebhookTriggerService as Service
        else:
            self.exit(1, 'Could not determine the service type.')

        return Service(self)

    @registry.factory(TcSession)
    @scoped_property
    def session_tc(self) -> 'TcSession':
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        return self.get_session_tc()

    @scoped_property
    def session_external(self) -> 'ExternalSession':
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        return self.get_session_external()

    def set_exit_code(self, exit_code: int):
        """Set the exit code (registry)"""
        self.exit_code = exit_code

    @registry.factory(Tokens, singleton=True)
    @cached_property
    def token(self) -> 'Tokens':
        """Return token object."""
        _proxies = None
        if self.inputs.model_unresolved.tc_proxy_tc is True:
            _proxies = self.proxies

        _tokens = Tokens(
            self.inputs.model_unresolved.tc_api_path,
            self.inputs.model_unresolved.tc_verify,
            _proxies,
        )

        # register token for Apps that pass token on start
        if all(
            [self.inputs.model_unresolved.tc_token, self.inputs.model_unresolved.tc_token_expires]
        ):
            _tokens.register_token(
                key=threading.current_thread().name,
                token=self.inputs.model_unresolved.tc_token,
                expires=self.inputs.model_unresolved.tc_token_expires,
            )
        return _tokens

    @property
    def ti_utils(self) -> 'ThreatIntelUtils':
        """Return instance of Threat Intel Utils."""
        return ThreatIntelUtils(self.session_tc)

    @cached_property
    def utils(self) -> 'Utils':  # pylint: disable=no-self-use
        """Include the Utils module."""
        return Utils()

    @cached_property
    def v2(self) -> 'V2':
        """Return a case management instance."""
        return V2(self.inputs, self.session_tc)

    @cached_property
    def v3(self) -> 'V3':
        """Return a case management instance."""
        return V3(self.session_tc)



"""Task / Tasks Object"""
# standard library
from typing import TYPE_CHECKING, Iterator, Union

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.tasks.task_filter import TaskFilter
from tcex.api.tc.v3.tasks.task_model import TaskModel, TasksModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact
    from tcex.api.tc.v3.notes.note import Note


class Tasks(ObjectCollectionABC):
    """Tasks Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = TasksModel(**kwargs)
        self.type_ = 'tasks'

    def __iter__(self) -> 'Task':
        """Iterate over CM objects."""
        return self.iterate(base_class=Task)

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TASKS.value

    @property
    def filter(self) -> 'TaskFilter':
        """Return the type specific filter object."""
        return TaskFilter(self.tql)


class Task(ObjectABC):
    """Tasks Object.

    Args:
        artifacts (Artifacts, kwargs): A list of Artifacts corresponding to the Task.
        assignee (Assignee, kwargs): The user or group Assignee object for the Task.
        case_id (int, kwargs): The **case id** for the Task.
        case_xid (str, kwargs): The **case xid** for the Task.
        completed_date (str, kwargs): The completion date of the Task.
        dependent_on_id (int, kwargs): The ID of another Task that this Task is dependent upon.
        description (str, kwargs): The **description** for the Task.
        due_date (str, kwargs): The due date of the Task.
        duration_type (str, kwargs): The **duration type** for the Task.
        name (str, kwargs): The **name** for the Task.
        notes (Notes, kwargs): A list of Notes corresponding to the Task.
        required (bool, kwargs): Flag indicating whether or not the task is required.
        status (str, kwargs): The **status** for the Task.
        workflow_phase (int, kwargs): The phase of the workflow.
        workflow_step (int, kwargs): The step of the workflow.
        xid (str, kwargs): The **xid** for the Task.
    """

    def __init__(self, **kwargs):
        """Initialize class properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model = TaskModel(**kwargs)
        self._nested_field_name = 'tasks'
        self._nested_filter = 'has_task'
        self.type_ = 'Task'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TASKS.value

    @property
    def model(self) -> 'TaskModel':
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: Union['TaskModel', dict]):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    @property
    def artifacts(self) -> Iterator['Artifact']:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)

    @property
    def notes(self) -> Iterator['Note']:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)

    def stage_artifact(self, data: Union[dict, 'ObjectABC', 'ArtifactModel']):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            raise RuntimeError('Invalid type passed in to stage_artifact')
        data._staged = True
        self.model.artifacts.data.append(data)

    # pylint: disable=redefined-builtin
    def stage_assignee(self, type: str, data: Union[dict, 'ObjectABC', 'ArtifactModel']):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif type.lower() == 'user' and isinstance(data, dict):
            data = UserModel(**data)
        elif type.lower() == 'group' and isinstance(data, dict):
            data = UserGroupModel(**data)

        if not isinstance(data, (UserModel, UserGroupModel)):
            raise RuntimeError('Invalid type passed in to stage_assignee')
        data._staged = True
        self.model.assignee._staged = True
        self.model.assignee.type = type
        self.model.assignee.data = data

    def stage_note(self, data: Union[dict, 'ObjectABC', 'NoteModel']):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            raise RuntimeError('Invalid type passed in to stage_note')
        data._staged = True
        self.model.notes.data.append(data)



"""Path Model"""
# standard library
import tempfile
from pathlib import Path

# third-party
from pydantic import BaseModel, Field


class PathModel(BaseModel):
    """Path Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * WebhookTriggerService
    * TriggerService
    """

    #
    # ThreatConnect Provided Inputs
    #

    tc_in_path: Path = Field(
        Path(tempfile.gettempdir() or '/tmp'),  # nosec
        description='The path to the Apps "in" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_log_path: Path = Field(
        Path(tempfile.gettempdir() or '/tmp'),  # nosec
        description='The path to the Apps "log" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_out_path: Path = Field(
        Path(tempfile.gettempdir() or '/tmp'),  # nosec
        description='The path to the Apps "out" directory.',
        inclusion_reason='runtimeLevel',
    )
    tc_temp_path: Path = Field(
        Path(tempfile.gettempdir() or '/tmp'),  # nosec
        description='The path to the Apps "tmp" directory.',
        inclusion_reason='runtimeLevel',
    )



"""ThreatConnect API V3 Types."""
# flake8: noqa
# standard library
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.adversary_assets.adversary_asset import AdversaryAsset
    from tcex.api.tc.v3.artifact_types.artifact_type import ArtifactType
    from tcex.api.tc.v3.artifacts.artifact import Artifact
    from tcex.api.tc.v3.cases.case import Case
    from tcex.api.tc.v3.groups.group import Group
    from tcex.api.tc.v3.indicators.indicator import Indicator
    from tcex.api.tc.v3.notes.note import Note
    from tcex.api.tc.v3.security.owner_roles.owner_role import OwnerRole
    from tcex.api.tc.v3.security.owners.owner import Owner
    from tcex.api.tc.v3.security.system_roles.system_role import SystemRole
    from tcex.api.tc.v3.security.user_groups.user_group import UserGroup
    from tcex.api.tc.v3.security.users.user import User
    from tcex.api.tc.v3.tags.tag import Tag
    from tcex.api.tc.v3.tasks.task import Task
    from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset
    from tcex.api.tc.v3.victims.victim import Victim
    from tcex.api.tc.v3.workflow_events.workflow_event import WorkflowEvent
    from tcex.api.tc.v3.workflow_templates.workflow_template import WorkflowTemplate

    V3Type = Union[
        AdversaryAsset,
        ArtifactType,
        Artifact,
        Case,
        Group,
        Indicator,
        Note,
        OwnerRole,
        Owner,
        SystemRole,
        UserGroup,
        User,
        Tag,
        Task,
        VictimAsset,
        Victim,
        WorkflowEvent,
        WorkflowTemplate,
    ]
